#!/usr/bin/env python
#
# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Test the CMX 3600 EDL adapter."""

# python
import os
import unittest

import opentimelineio as otio
import opentimelineio.test_utils as otio_test_utils
from opentimelineio.adapters import cmx_3600

# handle python2 vs python3 difference
try:
    from tempfile import TemporaryDirectory  # noqa: F401
    import tempfile
except ImportError:
    # XXX: python2.7 only
    from backports import tempfile


SAMPLE_DATA_DIR = os.path.join(os.path.dirname(__file__), "sample_data")
SCREENING_EXAMPLE_PATH = os.path.join(SAMPLE_DATA_DIR, "screening_example.edl")
NUCODA_EXAMPLE_PATH = os.path.join(SAMPLE_DATA_DIR, "nucoda_example.edl")
EXEMPLE_25_FPS_PATH = os.path.join(SAMPLE_DATA_DIR, "25fps.edl")
NO_SPACES_PATH = os.path.join(SAMPLE_DATA_DIR, "no_spaces_test.edl")
DISSOLVE_TEST = os.path.join(SAMPLE_DATA_DIR, "dissolve_test.edl")
DISSOLVE_TEST_2 = os.path.join(SAMPLE_DATA_DIR, "dissolve_test_2.edl")
DISSOLVE_TEST_3 = os.path.join(SAMPLE_DATA_DIR, "dissolve_test_3.edl")
DISSOLVE_TEST_4 = os.path.join(SAMPLE_DATA_DIR, "dissolve_test_4.edl")
GAP_TEST = os.path.join(SAMPLE_DATA_DIR, "gap_test.edl")
WIPE_TEST = os.path.join(SAMPLE_DATA_DIR, "wipe_test.edl")
TIMECODE_MISMATCH_TEST = os.path.join(SAMPLE_DATA_DIR, "timecode_mismatch.edl")
SPEED_EFFECTS_TEST = os.path.join(SAMPLE_DATA_DIR, "speed_effects.edl")
SPEED_EFFECTS_TEST_SMALL = os.path.join(
    SAMPLE_DATA_DIR,
    "speed_effects_small.edl"
)
MULTIPLE_TARGET_AUDIO_PATH = os.path.join(SAMPLE_DATA_DIR, "multi_audio.edl")
TRANSITION_DURATION_TEST = os.path.join(SAMPLE_DATA_DIR, "transition_duration.edl")
ENABLED_TEST = os.path.join(SAMPLE_DATA_DIR, "enabled.otio")


class EDLAdapterTest(unittest.TestCase, otio_test_utils.OTIOAssertions):
    maxDiff = None

    def test_edl_read(self):
        edl_path = SCREENING_EXAMPLE_PATH
        fps = 24
        timeline = otio.adapters.read_from_file(edl_path)
        self.assertTrue(timeline is not None)
        self.assertEqual(len(timeline.tracks), 1)
        self.assertEqual(len(timeline.tracks[0]), 9)
        self.assertEqual(
            timeline.tracks[0][0].name,
            "ZZ100_501 (LAY3)"
        )
        self.assertEqual(
            timeline.tracks[0][0].source_range.duration,
            otio.opentime.from_timecode("00:00:01:07", fps)
        )
        self.assertEqual(
            timeline.tracks[0][1].name,
            "ZZ100_502A (LAY3)"
        )
        self.assertEqual(
            timeline.tracks[0][1].source_range.duration,
            otio.opentime.from_timecode("00:00:02:02", fps)
        )
        self.assertEqual(
            timeline.tracks[0][2].name,
            "ZZ100_503A (LAY1)"
        )
        self.assertEqual(
            timeline.tracks[0][2].source_range.duration,
            otio.opentime.from_timecode("00:00:01:04", fps)
        )
        self.assertEqual(
            timeline.tracks[0][3].name,
            "ZZ100_504C (LAY1)"
        )
        self.assertEqual(
            timeline.tracks[0][3].source_range.duration,
            otio.opentime.from_timecode("00:00:04:19", fps)
        )

        self.assertEqual(len(timeline.tracks[0][3].markers), 2)
        marker = timeline.tracks[0][3].markers[0]
        self.assertEqual(marker.name, "ANIM FIX NEEDED")
        self.assertEqual(marker.metadata.get("cmx_3600").get("color"), "RED")
        self.assertEqual(
            marker.marked_range.start_time,
            otio.opentime.from_timecode("01:00:01:14", fps)
        )
        self.assertEqual(marker.color, otio.schema.MarkerColor.RED)

        unnamed_marker = timeline.tracks[0][6].markers[0]
        self.assertEqual(unnamed_marker.name, '')

        self.assertEqual(
            timeline.tracks[0][4].name,
            "ZZ100_504B (LAY1)"
        )
        self.assertEqual(
            timeline.tracks[0][4].source_range.duration,
            otio.opentime.from_timecode("00:00:04:05", fps)
        )
        self.assertEqual(
            timeline.tracks[0][5].name,
            "ZZ100_507C (LAY2)"
        )
        self.assertEqual(
            timeline.tracks[0][5].source_range.duration,
            otio.opentime.from_timecode("00:00:06:17", fps)
        )
        self.assertEqual(
            timeline.tracks[0][6].name,
            "ZZ100_508 (LAY2)"
        )
        self.assertEqual(
            timeline.tracks[0][6].source_range.duration,
            otio.opentime.from_timecode("00:00:07:02", fps)
        )
        self.assertEqual(
            timeline.tracks[0][7].name,
            "ZZ100_510 (LAY1)"
        )
        self.assertEqual(
            timeline.tracks[0][7].source_range.duration,
            otio.opentime.from_timecode("00:00:05:16", fps)
        )
        self.assertEqual(
            timeline.tracks[0][8].name,
            "ZZ100_510B (LAY1)"
        )
        self.assertEqual(
            timeline.tracks[0][8].source_range.duration,
            otio.opentime.from_timecode("00:00:10:17", fps)
        )

    def test_reelname_length(self):
        track = otio.schema.Track()
        tl = otio.schema.Timeline("test_timeline", tracks=[track])
        rt = otio.opentime.RationalTime(5.0, 24.0)

        long_mr = otio.schema.ExternalReference(
            target_url="/var/tmp/test_a_really_really_long_filename.mov"
        )

        tr = otio.opentime.TimeRange(
            start_time=otio.opentime.RationalTime(0.0, 24.0),
            duration=rt
        )

        cl = otio.schema.Clip(
            name="test clip1",
            media_reference=long_mr,
            source_range=tr,
        )

        track.name = "V1"
        track.append(cl)

        # Test default behavior
        result = otio.adapters.write_to_string(tl, adapter_name="cmx_3600")

        expected = '''TITLE: test_timeline

001  testarea V     C        00:00:00:00 00:00:00:05 00:00:00:00 00:00:00:05
* FROM CLIP NAME:  test clip1
* FROM CLIP: /var/tmp/test_a_really_really_long_filename.mov
* OTIO TRUNCATED REEL NAME FROM: test_a_really_really_long_filename.mov
'''

        self.assertMultiLineEqual(result, expected)

        # Keep full filename (minus extension) as reelname
        result = otio.adapters.write_to_string(
            tl,
            adapter_name="cmx_3600",
            reelname_len=None
        )
        expected = '''TITLE: test_timeline

001  test_a_really_really_long_filename \
V     C        00:00:00:00 00:00:00:05 00:00:00:00 00:00:00:05
* FROM CLIP NAME:  test clip1
* FROM CLIP: /var/tmp/test_a_really_really_long_filename.mov
'''

        self.assertMultiLineEqual(result, expected)

        # Make sure reel name is only 12 characters long
        result = otio.adapters.write_to_string(
            tl,
            adapter_name="cmx_3600",
            reelname_len=12
        )
        expected = '''TITLE: test_timeline

001  testareallyr V     C        00:00:00:00 00:00:00:05 00:00:00:00 00:00:00:05
* FROM CLIP NAME:  test clip1
* FROM CLIP: /var/tmp/test_a_really_really_long_filename.mov
* OTIO TRUNCATED REEL NAME FROM: test_a_really_really_long_filename.mov
'''

        self.assertMultiLineEqual(result, expected)

    def test_edl_round_trip_mem2disk2mem(self):
        track = otio.schema.Track()
        tl = otio.schema.Timeline("test_timeline", tracks=[track])
        rt = otio.opentime.RationalTime(5.0, 24.0)
        mr = otio.schema.ExternalReference(target_url="/var/tmp/test.mov")
        md = {
            "cmx_3600": {
                "reel": "test",
                "comments": ["OTIO TRUNCATED REEL NAME FROM: test.mov"]
            }
        }

        tr = otio.opentime.TimeRange(
            start_time=otio.opentime.RationalTime(0.0, 24.0),
            duration=rt
        )

        cl = otio.schema.Clip(
            name="test clip1",
            media_reference=mr,
            source_range=tr,
            metadata=md
        )
        cl2 = otio.schema.Clip(
            name="test clip2",
            media_reference=mr.clone(),
            source_range=tr,
            metadata=md
        )
        cl3 = otio.schema.Clip(
            name="test clip3",
            media_reference=mr.clone(),
            source_range=tr,
            metadata=md
        )
        cl4 = otio.schema.Clip(
            name="test clip3_ff",
            media_reference=mr.clone(),
            source_range=tr,
            metadata=md
        )

        cl4.effects[:] = [otio.schema.FreezeFrame()]
        cl5 = otio.schema.Clip(
            name="test clip5 (speed)",
            media_reference=mr.clone(),
            source_range=tr,
            metadata=md
        )
        cl5.effects[:] = [otio.schema.LinearTimeWarp(time_scalar=2.0)]
        track.name = "V"
        track.append(cl)
        track.extend([cl2, cl3])
        track.append(cl4)
        track.append(cl5)

        result = otio.adapters.write_to_string(tl, adapter_name="cmx_3600")
        new_otio = otio.adapters.read_from_string(
            result,
            adapter_name="cmx_3600"
        )

        # directly compare clip with speed effect
        self.assertEqual(
            len(new_otio.tracks[0][3].effects),
            1
        )
        self.assertEqual(
            new_otio.tracks[0][3].name,
            tl.tracks[0][3].name
        )

        self.assertJsonEqual(new_otio, tl)

        # ensure that an error is raised if more than one effect is present
        cl5.effects.append(otio.schema.FreezeFrame())
        with self.assertRaises(otio.exceptions.NotSupportedError):
            otio.adapters.write_to_string(tl, "cmx_3600")

        # blank effect should pass through and be ignored
        cl5.effects[:] = [otio.schema.Effect()]
        otio.adapters.write_to_string(tl, "cmx_3600")

        # but a timing effect should raise an exception
        cl5.effects[:] = [otio.schema.TimeEffect()]
        with self.assertRaises(otio.exceptions.NotSupportedError):
            otio.adapters.write_to_string(tl, "cmx_3600")

    def test_edl_round_trip_disk2mem2disk_speed_effects(self):
        test_edl = SPEED_EFFECTS_TEST_SMALL
        timeline = otio.adapters.read_from_file(test_edl)

        with tempfile.TemporaryDirectory() as temp_dir:
            tmp_path = os.path.join(
                temp_dir,
                "test_edl_round_trip_disk2mem2disk_speed_effects.edl"
            )

            otio.adapters.write_to_file(timeline, tmp_path)

            result = otio.adapters.read_from_file(tmp_path)

            # When debugging, you can use this to see the difference in the OTIO
            # otio.adapters.otio_json.write_to_file(timeline, "/tmp/original.otio")
            # otio.adapters.otio_json.write_to_file(result, "/tmp/output.otio")
            # os.system("xxdiff /tmp/{original,output}.otio")

            # When debugging, use this to see the difference in the EDLs on disk
            # os.system("xxdiff {} {}&".format(test_edl, tmp_path))

            # The in-memory OTIO representation should be the same
            self.assertJsonEqual(timeline, result)

    def test_edl_round_trip_disk2mem2disk(self):
        timeline = otio.adapters.read_from_file(SCREENING_EXAMPLE_PATH)

        with tempfile.TemporaryDirectory() as temp_dir:
            tmp_path = os.path.join(
                temp_dir,
                "test_edl_round_trip_disk2mem2disk.edl"
            )

            otio.adapters.write_to_file(timeline, tmp_path)

            result = otio.adapters.read_from_file(tmp_path)

            # When debugging, you can use this to see the difference in the OTIO
            # otio.adapters.otio_json.write_to_file(timeline, "/tmp/original.otio")
            # otio.adapters.otio_json.write_to_file(result, "/tmp/output.otio")
            # os.system("opendiff /tmp/{original,output}.otio")

            original_json = otio.adapters.otio_json.write_to_string(timeline)
            output_json = otio.adapters.otio_json.write_to_string(result)
            self.assertMultiLineEqual(original_json, output_json)

            # The in-memory OTIO representation should be the same
            self.assertIsOTIOEquivalentTo(timeline, result)

            # When debugging, use this to see the difference in the EDLs on disk
            # os.system("opendiff {} {}".format(SCREENING_EXAMPLE_PATH, tmp_path))

            # But the EDL text on disk are *not* byte-for-byte identical
            with open(SCREENING_EXAMPLE_PATH, "r") as original_file:
                with open(tmp_path, "r") as output_file:
                    self.assertNotEqual(original_file.read(), output_file.read())

    def test_regex_flexibility(self):
        timeline = otio.adapters.read_from_file(SCREENING_EXAMPLE_PATH)
        no_spaces = otio.adapters.read_from_file(NO_SPACES_PATH)
        self.assertIsOTIOEquivalentTo(timeline, no_spaces)

    def test_clip_with_tab_and_space_delimiters(self):
        timeline = otio.adapters.read_from_string(
            '001  Z10 V  C\t\t01:00:04:05 01:00:05:12 00:59:53:11 00:59:54:18',
            adapter_name="cmx_3600"
        )
        self.assertTrue(timeline is not None)
        self.assertEqual(len(timeline.tracks), 1)
        self.assertEqual(
            timeline.tracks[0].kind,
            otio.schema.TrackKind.Video
        )
        self.assertEqual(len(timeline.tracks[0]), 1)
        self.assertEqual(
            timeline.tracks[0][0].source_range.start_time.value,
            86501
        )
        self.assertEqual(
            timeline.tracks[0][0].source_range.duration.value,
            31
        )

    def test_imagesequence_read(self):
        trunced_edl1 = '''TITLE: Image Sequence Write

001  myimages V     C        01:00:01:00 01:00:02:12 00:00:00:00 00:00:01:12
* FROM CLIP NAME:  my_image_sequence
* FROM CLIP: /media/path/my_image_sequence.[1025-1060].ext
* OTIO TRUNCATED REEL NAME FROM: my_image_sequence.[1025-1060].ext
'''
        rate = 24
        tl1 = otio.adapters.read_from_string(trunced_edl1, 'cmx_3600', rate=rate)
        self.assertIsInstance(tl1, otio.schema.Timeline)

        clip1 = tl1.tracks[0][0]
        media_ref1 = clip1.media_reference
        self.assertIsInstance(media_ref1, otio.schema.ImageSequenceReference)
        self.assertEqual(media_ref1.start_frame, 1025)
        self.assertEqual(media_ref1.end_frame(), 1060)
        self.assertEqual(
            clip1.available_range(),
            otio.opentime.range_from_start_end_time(
                otio.opentime.from_timecode('01:00:01:00', rate),
                otio.opentime.from_timecode('01:00:02:12', rate)
            )
        )

        # Make sure regex works and uses ExternalReference for non sequences
        trunced_edl2 = '''TITLE: Image Sequence Write

001  myimages V     C        01:00:01:00 01:00:02:12 00:00:00:00 00:00:01:12
* FROM CLIP NAME:  my_image_sequence
* FROM CLIP: /media/path/my_image_file.1025.ext
* OTIO TRUNCATED REEL NAME FROM: my_image_file.1025.ext
'''

        tl2 = otio.adapters.read_from_string(trunced_edl2, 'cmx_3600', rate=rate)
        clip2 = tl2.tracks[0][0]
        media_ref2 = clip2.media_reference
        self.assertIsInstance(media_ref2, otio.schema.ExternalReference)

        trunced_edl3 = '''TITLE: Image Sequence Write

001  myimages V     C        01:00:01:00 01:00:02:12 00:00:00:00 00:00:01:12
* FROM CLIP NAME:  my_image_sequence
* FROM CLIP: /media/path/my_image_file.[1025].ext
* OTIO TRUNCATED REEL NAME FROM: my_image_file.[1025].ext
'''
        tl3 = otio.adapters.read_from_string(trunced_edl3, 'cmx_3600', rate=rate)
        clip3 = tl3.tracks[0][0]
        media_ref3 = clip3.media_reference
        self.assertIsInstance(media_ref3, otio.schema.ExternalReference)

    def test_imagesequence_write(self):
        rate = 24
        tl = otio.schema.Timeline('Image Sequence Write')
        track = otio.schema.Track('V1')
        tl.tracks.append(track)

        clip = otio.schema.Clip(
            name='my_image_sequence',
            source_range=otio.opentime.range_from_start_end_time(
                otio.opentime.from_timecode('01:00:01:00', rate),
                otio.opentime.from_timecode('01:00:02:12', rate)
            ),
            media_reference=otio.schema.ImageSequenceReference(
                target_url_base='/media/path/',
                name_prefix='my_image_sequence.',
                name_suffix='.ext',
                rate=rate,
                start_frame=1001,
                frame_zero_padding=4,
                available_range=otio.opentime.range_from_start_end_time(
                    otio.opentime.from_timecode('01:00:00:00', rate),
                    otio.opentime.from_timecode('01:00:03:00', rate)
                )
            )
        )
        track.append(clip)

        # Default behavior
        result1 = otio.adapters.write_to_string(tl, 'cmx_3600', rate=rate)

        expected_result1 = '''TITLE: Image Sequence Write

001  myimages V     C        01:00:01:00 01:00:02:12 00:00:00:00 00:00:01:12
* FROM CLIP NAME:  my_image_sequence
* FROM CLIP: /media/path/my_image_sequence.[1025-1060].ext
* OTIO TRUNCATED REEL NAME FROM: my_image_sequence.[1025-1060].ext
'''
        self.assertMultiLineEqual(result1, expected_result1)

        # Only trunc extension in reel name
        result2 = otio.adapters.write_to_string(
            tl,
            'cmx_3600',
            rate=24,
            reelname_len=None
        )

        expected_result2 = '''TITLE: Image Sequence Write

001  my_image_sequence.[1025-1060] V     C        \
01:00:01:00 01:00:02:12 00:00:00:00 00:00:01:12
* FROM CLIP NAME:  my_image_sequence
* FROM CLIP: /media/path/my_image_sequence.[1025-1060].ext
'''
        self.assertMultiLineEqual(result2, expected_result2)

    def test_dissolve_parse(self):
        tl = otio.adapters.read_from_file(DISSOLVE_TEST)
        # clip/transition/clip/clip
        self.assertEqual(len(tl.tracks[0]), 4)

        self.assertTrue(isinstance(tl.tracks[0][1], otio.schema.Transition))
        self.assertEqual(tl.tracks[0][0].duration().value, 9)
        # The visible range must contains all the frames needed for the transition
        # Edit duration + transition duration
        self.assertEqual(tl.tracks[0][0].visible_range().duration.to_frames(), 19)
        self.assertEqual(tl.tracks[0][0].name, "clip_A")
        self.assertEqual(tl.tracks[0][1].duration().value, 10)
        self.assertEqual(tl.tracks[0][1].name, "SMPTE_Dissolve from clip_A to clip_B")
        self.assertEqual(tl.tracks[0][2].duration().value, 10)
        self.assertEqual(tl.tracks[0][2].visible_range().duration.value, 10)
        self.assertEqual(tl.tracks[0][2].name, "clip_B")
        self.assertEqual(tl.tracks[0][3].duration().value, 1)
        self.assertEqual(tl.tracks[0][2].name, "clip_B")

    def test_dissolve_parse_middle(self):
        tl = otio.adapters.read_from_file(DISSOLVE_TEST_2)
        trck = tl.tracks[0]
        # 3 clips and 1 transition
        self.assertEqual(len(trck), 4)

        self.assertTrue(isinstance(trck[1], otio.schema.Transition))

        self.assertEqual(trck[0].duration().value, 5)
        self.assertEqual(trck[0].visible_range().duration.to_frames(), 15)
        self.assertEqual(trck[1].duration().value, 10)
        self.assertEqual(trck[1].name, "SMPTE_Dissolve from clip_A to clip_B")

        self.assertEqual(
            trck[2].source_range.start_time.value,
            otio.opentime.from_timecode('01:00:08:04', 24).value
        )
        self.assertEqual(trck[2].name, "clip_B")
        self.assertEqual(trck[2].duration().value, 10)
        self.assertEqual(trck[2].visible_range().duration.value, 10)

        self.assertEqual(tl.tracks[0][0].visible_range().duration.to_frames(), 15)

    def test_dissolve_parse_full_clip_dissolve(self):
        tl = otio.adapters.read_from_file(DISSOLVE_TEST_3)
        self.assertEqual(len(tl.tracks[0]), 4)

        self.assertTrue(isinstance(tl.tracks[0][1], otio.schema.Transition))

        trck = tl.tracks[0]
        clip_a = trck[0]
        self.assertEqual(clip_a.name, "Clip_A.mov")
        self.assertEqual(clip_a.duration().value, 61)
        self.assertEqual(clip_a.visible_range().duration.value, 61 + 30)

        transition = trck[1]
        # Note: clip names in the EDL are wrong, the transition is actually
        # from Clip_A to Clip_B
        self.assertEqual(
            transition.name,
            "SMPTE_Dissolve from Clip_B.mov to Clip_C.mov"
        )
        self.assertEqual(transition.in_offset.value, 0)
        self.assertEqual(transition.out_offset.value, 30)

        clip_c = trck[2]
        self.assertEqual(clip_c.name, "Clip_C.mov")
        self.assertEqual(clip_c.source_range.start_time.value, 86400 + (33 * 24 + 22))
        self.assertEqual(clip_c.duration().value, 30)
        self.assertEqual(clip_c.visible_range().duration.value, 30)

        clip_d = trck[3]
        self.assertEqual(clip_d.name, "Clip_D.mov")
        self.assertEqual(clip_d.source_range.start_time.value, 86400)
        self.assertEqual(clip_d.duration().value, 46)

    def test_dissolve_with_odd_frame_count_maintains_length(self):
        # EXERCISE
        tl = otio.adapters.read_from_string(
            '1 CLPA V C     00:00:04:17 00:00:07:02 00:00:00:00 00:00:02:09\n'
            '2 CLPA V C     00:00:07:02 00:00:07:02 00:00:02:09 00:00:02:09\n'
            '2 CLPB V D 027 00:00:06:18 00:00:07:21 00:00:02:09 00:00:03:12\n'
            '3 CLPB V C     00:00:07:21 00:00:15:21 00:00:03:12 00:00:11:12\n',
            adapter_name="cmx_3600"
        )

        # VALIDATE
        self.assertEqual(tl.duration().value, (11 * 24) + 12)

    def test_wipe_parse(self):
        tl = otio.adapters.read_from_file(WIPE_TEST)
        self.assertEqual(len(tl.tracks[0]), 4)

        wipe = tl.tracks[0][1]
        self.assertTrue(isinstance(wipe, otio.schema.Transition))
        self.assertEqual(wipe.transition_type, "SMPTE_Wipe")
        self.assertEqual(wipe.metadata["cmx_3600"]["transition"], "W001")

        self.assertEqual(tl.tracks[0][0].duration().value, 9)
        self.assertEqual(tl.tracks[0][0].visible_range().duration.value, 19)

        self.assertEqual(tl.tracks[0][2].duration().value, 10)
        self.assertEqual(tl.tracks[0][2].visible_range().duration.value, 10)

        self.assertEqual(tl.tracks[0][3].duration().value, 1)

    def test_fade_to_black(self):
        # EXERCISE
        tl = otio.adapters.read_from_string(
            '1 CLPA V C     00:00:03:18 00:00:12:15 00:00:00:00 00:00:08:21\n'
            '2 CLPA V C     00:00:12:15 00:00:12:15 00:00:08:21 00:00:08:21\n'
            '2 BL   V D 024 00:00:00:00 00:00:01:00 00:00:08:21 00:00:09:21\n',
            adapter_name="cmx_3600"
        )

        # VALIDATE
        self.assertEqual(len(tl.tracks[0]), 3)
        self.assertTrue(isinstance(tl.tracks[0][1], otio.schema.Transition))
        self.assertTrue(isinstance(tl.tracks[0][2], otio.schema.Clip))
        self.assertEqual(tl.tracks[0][2].media_reference.generator_kind, 'black')
        self.assertEqual(tl.tracks[0][2].duration().value, 24)
        self.assertEqual(tl.tracks[0][2].source_range.start_time.value, 0)

    def test_edl_round_trip_with_transitions(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            # Notes:
            # - the writer does not handle wipes, only dissolves
            # - the writer can generate invalid EDLs if spaces are in reel names.
            for edl_file in [
                DISSOLVE_TEST,
                DISSOLVE_TEST_2,
                DISSOLVE_TEST_3,
                DISSOLVE_TEST_4
            ]:
                edl_name = os.path.basename(edl_file)
                timeline = otio.adapters.read_from_file(edl_file)
                tmp_path = os.path.join(
                    temp_dir,
                    'test_edl_round_trip_{}'.format(edl_name)
                )
                otio.adapters.write_to_file(timeline, tmp_path)

                result = otio.adapters.read_from_file(tmp_path)
                self.assertEqual(len(timeline.tracks), len(result.tracks))
                for track, res_track in zip(timeline.tracks, result.tracks):
                    self.assertEqual(len(track), len(res_track))
                    for child, res_child in zip(track, res_track):
                        self.assertEqual(type(child), type(res_child))
                        if isinstance(child, otio.schema.Transition):
                            self.assertEqual(child.in_offset, res_child.in_offset)
                            self.assertEqual(child.out_offset, res_child.out_offset)
                            self.assertEqual(
                                child.transition_type, res_child.transition_type
                            )
                        else:
                            self.assertEqual(child.source_range, res_child.source_range)

    def test_edl_25fps(self):
        # EXERCISE
        edl_path = EXEMPLE_25_FPS_PATH
        fps = 25
        timeline = otio.adapters.read_from_file(edl_path, rate=fps)
        track = timeline.tracks[0]
        self.assertEqual(track[0].source_range.duration.value, 161)
        self.assertEqual(track[1].source_range.duration.value, 200)
        self.assertEqual(track[2].source_range.duration.value, 86)
        self.assertEqual(track[3].source_range.duration.value, 49)

    def test_record_gaps(self):
        edl_path = GAP_TEST
        timeline = otio.adapters.read_from_file(edl_path)
        track = timeline.tracks[0]
        self.assertEqual(len(track), 5)
        self.assertEqual(track.duration().value, 5 * 24 + 6)
        clip1, gapA, clip2, gapB, clip3 = track[:]
        self.assertEqual(clip1.source_range.duration.value, 24)
        self.assertEqual(clip2.source_range.duration.value, 24)
        self.assertEqual(clip3.source_range.duration.value, 24)
        self.assertEqual(gapA.duration().value, 16)
        self.assertEqual(gapB.duration().value, 38)
        self.assertEqual(clip1.range_in_parent().duration.value, 24)
        self.assertEqual(clip2.range_in_parent().duration.value, 24)
        self.assertEqual(clip3.range_in_parent().duration.value, 24)
        self.assertEqual(
            [item.range_in_parent() for item in track],
            [
                otio.opentime.TimeRange(
                    otio.opentime.from_frames(0, 24),
                    otio.opentime.from_frames(24, 24)
                ),
                otio.opentime.TimeRange(
                    otio.opentime.from_frames(24, 24),
                    otio.opentime.from_frames(16, 24)
                ),
                otio.opentime.TimeRange(
                    otio.opentime.from_frames(40, 24),
                    otio.opentime.from_frames(24, 24)
                ),
                otio.opentime.TimeRange(
                    otio.opentime.from_frames(64, 24),
                    otio.opentime.from_frames(38, 24)
                ),
                otio.opentime.TimeRange(
                    otio.opentime.from_frames(102, 24),
                    otio.opentime.from_frames(24, 24)
                )
            ]
        )

    def test_read_generators(self):
        # EXERCISE
        tl = otio.adapters.read_from_string(
            '1 BL V C 00:00:00:00 00:00:01:00 00:00:00:00 00:00:01:00\n'
            '2 BLACK V C 00:00:00:00 00:00:01:00 00:00:01:00 00:00:02:00\n'
            '3 BARS V C 00:00:00:00 00:00:01:00 00:00:02:00 00:00:03:00\n',
            adapter_name="cmx_3600"
        )

        # VALIDATE
        self.assertEqual(
            tl.tracks[0][0].media_reference.generator_kind,
            'black'
        )
        self.assertEqual(
            tl.tracks[0][1].media_reference.generator_kind,
            'black'
        )
        self.assertEqual(
            tl.tracks[0][2].media_reference.generator_kind,
            'SMPTEBars'
        )

    def test_nucoda_edl_read(self):
        edl_path = NUCODA_EXAMPLE_PATH
        fps = 24
        timeline = otio.adapters.read_from_file(edl_path)
        self.assertTrue(timeline is not None)
        self.assertEqual(len(timeline.tracks), 1)
        self.assertEqual(len(timeline.tracks[0]), 2)
        self.assertEqual(
            timeline.tracks[0][0].name,
            "take_1"
        )
        self.assertEqual(
            timeline.tracks[0][0].source_range.duration,
            otio.opentime.from_timecode("00:00:01:07", fps)
        )
        self.assertIsOTIOEquivalentTo(
            timeline.tracks[0][0].media_reference,
            otio.schema.ExternalReference(
                target_url=r"S:\path\to\ZZ100_501.take_1.0001.exr"
            )
        )
        self.assertEqual(
            timeline.tracks[0][1].name,
            "take_2"
        )
        self.assertEqual(
            timeline.tracks[0][1].source_range.duration,
            otio.opentime.from_timecode("00:00:02:02", fps)
        )
        self.assertIsOTIOEquivalentTo(
            timeline.tracks[0][1].media_reference,
            otio.schema.ExternalReference(
                target_url=r"S:\path\to\ZZ100_502A.take_2.0101.exr"
            )
        )

    def test_nucoda_edl_write(self):
        track = otio.schema.Track()
        tl = otio.schema.Timeline("test_nucoda_timeline", tracks=[track])
        rt = otio.opentime.RationalTime(5.0, 24.0)
        mr = otio.schema.ExternalReference(target_url=r"S:\var\tmp\test.exr")

        tr = otio.opentime.TimeRange(
            start_time=otio.opentime.RationalTime(0.0, 24.0),
            duration=rt
        )
        cl = otio.schema.Clip(
            name="test clip1",
            media_reference=mr,
            source_range=tr,
        )
        gap = otio.schema.Gap(
            source_range=otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(0, 24.0),
                duration=otio.opentime.RationalTime(24.0, 24.0),
            )
        )
        cl2 = otio.schema.Clip(
            name="test clip2",
            media_reference=mr.clone(),
            source_range=tr,
        )
        tl.tracks[0].name = "V"
        tl.tracks[0].append(cl)
        tl.tracks[0].append(gap)
        tl.tracks[0].append(cl2)

        result = otio.adapters.write_to_string(
            tl,
            adapter_name='cmx_3600',
            style='nucoda'
        )

        expected = r'''TITLE: test_nucoda_timeline

001  test     V     C        00:00:00:00 00:00:00:05 00:00:00:00 00:00:00:05
* FROM CLIP NAME:  test clip1
* FROM FILE: S:\var\tmp\test.exr
* OTIO TRUNCATED REEL NAME FROM: test.exr
002  test     V     C        00:00:00:00 00:00:00:05 00:00:01:05 00:00:01:10
* FROM CLIP NAME:  test clip2
* FROM FILE: S:\var\tmp\test.exr
* OTIO TRUNCATED REEL NAME FROM: test.exr
'''

        self.assertMultiLineEqual(result, expected)

    def test_reels_edl_round_trip_string2mem2string(self):

        sample_data = r'''TITLE: Reels_Example.01

001  ZZ100_50 V     C        01:00:04:05 01:00:05:12 00:59:53:11 00:59:54:18
* FROM CLIP NAME:  take_1
* FROM FILE: S:\path\to\ZZ100_501.take_1.0001.exr
002  ZZ100_50 V     C        01:00:06:13 01:00:08:15 00:59:54:18 00:59:56:20
* FROM CLIP NAME:  take_2
* FROM FILE: S:\path\to\ZZ100_502A.take_2.0101.exr
'''

        timeline = otio.adapters.read_from_string(sample_data, adapter_name="cmx_3600")
        otio_data = otio.adapters.write_to_string(timeline, adapter_name="cmx_3600",
                                                  style="nucoda")
        self.assertMultiLineEqual(sample_data, otio_data)

    def test_nucoda_edl_write_with_transition(self):
        track = otio.schema.Track()
        tl = otio.schema.Timeline(
            "Example CrossDissolve",
            tracks=[track]
        )

        cl = otio.schema.Clip(
            'Clip1',
            metadata={'cmx_3600': {'reel': 'Clip1'}},
            media_reference=otio.schema.ExternalReference(
                target_url="/var/tmp/clip1.001.exr"
            ),
            source_range=otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(131.0, 24.0),
                duration=otio.opentime.RationalTime(102.0, 24.0)
            )
        )
        trans = otio.schema.Transition(
            in_offset=otio.opentime.RationalTime(57.0, 24.0),
            out_offset=otio.opentime.RationalTime(43.0, 24.0)
        )
        cl2 = otio.schema.Clip(
            'Clip2',
            metadata={'cmx_3600': {'reel': 'Clip2'}},
            media_reference=otio.schema.ExternalReference(
                target_url="/var/tmp/clip2.001.exr"
            ),
            source_range=otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(280.0, 24.0),
                duration=otio.opentime.RationalTime(143.0, 24.0)
            )
        )
        cl3 = otio.schema.Clip(
            'Clip3',
            metadata={'cmx_3600': {'reel': 'Clip3'}},
            media_reference=otio.schema.ExternalReference(
                target_url="/var/tmp/clip3.001.exr"
            ),
            source_range=otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(0.0, 24.0),
                duration=otio.opentime.RationalTime(24.0, 24.0)
            )
        )
        tl.tracks[0].extend([cl, trans, cl2, cl3])

        result = otio.adapters.write_to_string(
            tl,
            adapter_name='cmx_3600',
            style='nucoda'
        )

        expected = r'''TITLE: Example CrossDissolve

001  Clip1    V     C        00:00:05:11 00:00:07:08 00:00:00:00 00:00:01:21
* FROM CLIP NAME:  Clip1
* FROM FILE: /var/tmp/clip1.001.exr
002  Clip1    V     C        00:00:07:08 00:00:07:08 00:00:01:21 00:00:01:21
002  Clip2    V     D 100    00:00:09:07 00:00:17:15 00:00:01:21 00:00:10:05
* FROM CLIP NAME:  Clip1
* FROM FILE: /var/tmp/clip1.001.exr
* TO CLIP NAME:  Clip2
* TO FILE: /var/tmp/clip2.001.exr
003  Clip3    V     C        00:00:00:00 00:00:01:00 00:00:10:05 00:00:11:05
* FROM CLIP NAME:  Clip3
* FROM FILE: /var/tmp/clip3.001.exr
'''

        self.assertMultiLineEqual(result, expected)

    def test_nucoda_edl_write_fade_in(self):
        track = otio.schema.Track()
        tl = otio.schema.Timeline(
            "Example Fade In",
            tracks=[track]
        )

        trans = otio.schema.Transition(
            in_offset=otio.opentime.RationalTime(0.0, 24.0),
            out_offset=otio.opentime.RationalTime(12.0, 24.0)
        )
        cl = otio.schema.Clip(
            'My Clip',
            metadata={'cmx_3600': {'reel': 'My_Clip'}},
            media_reference=otio.schema.ExternalReference(
                target_url="/var/tmp/clip.001.exr"
            ),
            source_range=otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(50.0, 24.0),
                duration=otio.opentime.RationalTime(26.0, 24.0)
            )
        )
        tl.tracks[0].extend([trans, cl])

        result = otio.adapters.write_to_string(
            tl,
            adapter_name='cmx_3600',
            style='nucoda'
        )

        expected = r'''TITLE: Example Fade In

001  BL       V     C        00:00:00:00 00:00:00:00 00:00:00:00 00:00:00:00
001  My_Clip  V     D 012    00:00:02:02 00:00:03:04 00:00:00:00 00:00:01:02
* TO CLIP NAME:  My Clip
* TO FILE: /var/tmp/clip.001.exr
'''

        self.assertMultiLineEqual(result, expected)

    def test_nucoda_edl_write_fade_out(self):
        track = otio.schema.Track()
        tl = otio.schema.Timeline(
            "Example Fade Out",
            tracks=[track]
        )

        cl = otio.schema.Clip(
            'My Clip',
            metadata={'cmx_3600': {'reel': 'My_Clip'}},
            media_reference=otio.schema.ExternalReference(
                target_url="/var/tmp/clip.001.exr"
            ),
            source_range=otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(24.0, 24.0),
                duration=otio.opentime.RationalTime(24.0, 24.0)
            )
        )
        trans = otio.schema.Transition(
            in_offset=otio.opentime.RationalTime(12.0, 24.0),
            out_offset=otio.opentime.RationalTime(0.0, 24.0)
        )
        tl.tracks[0].extend([cl, trans])

        result = otio.adapters.write_to_string(
            tl,
            adapter_name='cmx_3600',
            style='nucoda'
        )

        expected = r'''TITLE: Example Fade Out

001  My_Clip  V     C        00:00:01:00 00:00:01:12 00:00:00:00 00:00:00:12
* FROM CLIP NAME:  My Clip
* FROM FILE: /var/tmp/clip.001.exr
002  My_Clip  V     C        00:00:01:12 00:00:01:12 00:00:00:12 00:00:00:12
002  BL       V     D 012    00:00:00:00 00:00:00:12 00:00:00:12 00:00:01:00
* FROM CLIP NAME:  My Clip
* FROM FILE: /var/tmp/clip.001.exr
'''

        self.assertMultiLineEqual(result, expected)

    def test_nucoda_edl_write_with_double_transition(self):
        track = otio.schema.Track()
        tl = otio.schema.Timeline("Double Transition", tracks=[track])

        cl = otio.schema.Clip(
            metadata={'cmx_3600': {'reel': 'Reel1'}},
            source_range=otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(24.0, 24.0),
                duration=otio.opentime.RationalTime(24.0, 24.0)
            )
        )
        trans = otio.schema.Transition(
            in_offset=otio.opentime.RationalTime(6.0, 24.0),
            out_offset=otio.opentime.RationalTime(6.0, 24.0)
        )
        cl2 = otio.schema.Clip(
            metadata={'cmx_3600': {'reel': 'Reel2'}},
            source_range=otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(24.0, 24.0),
                duration=otio.opentime.RationalTime(24.0, 24.0)
            )
        )
        trans2 = otio.schema.Transition(
            in_offset=otio.opentime.RationalTime(6.0, 24.0),
            out_offset=otio.opentime.RationalTime(6.0, 24.0)
        )
        cl3 = otio.schema.Clip(
            metadata={'cmx_3600': {'reel': 'Reel3'}},
            source_range=otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(24.0, 24.0),
                duration=otio.opentime.RationalTime(24.0, 24.0)
            )
        )
        tl.tracks[0].extend([cl, trans, cl2, trans2, cl3])

        result = otio.adapters.write_to_string(
            tl,
            adapter_name='cmx_3600',
            style='nucoda'
        )

        expected = '''TITLE: Double Transition

001  Reel1    V     C        00:00:01:00 00:00:01:18 00:00:00:00 00:00:00:18
002  Reel1    V     C        00:00:01:18 00:00:01:18 00:00:00:18 00:00:00:18
002  Reel2    V     D 012    00:00:00:18 00:00:01:18 00:00:00:18 00:00:01:18
003  Reel2    V     C        00:00:01:18 00:00:01:18 00:00:01:18 00:00:01:18
003  Reel3    V     D 012    00:00:00:18 00:00:02:00 00:00:01:18 00:00:03:00
'''

        self.assertMultiLineEqual(result, expected)

    def test_read_edl_with_multiple_target_audio_tracks(self):
        tl = otio.adapters.read_from_file(MULTIPLE_TARGET_AUDIO_PATH)

        self.assertEqual(len(tl.audio_tracks()), 2)

        first_track, second_track = tl.audio_tracks()
        self.assertEqual(first_track.name, "A1")
        self.assertEqual(second_track.name, "A2")

        self.assertEqual(first_track[0].name, "AX")
        self.assertEqual(second_track[0].name, "AX")

        expected_range = otio.opentime.TimeRange(
            duration=otio.opentime.from_timecode("00:56:55:22", rate=24)
        )
        self.assertEqual(first_track[0].source_range, expected_range)
        self.assertEqual(second_track[0].source_range, expected_range)

    def test_custom_reel_names(self):
        track = otio.schema.Track()
        tl = otio.schema.Timeline(tracks=[track])
        tr = otio.opentime.TimeRange(
            start_time=otio.opentime.RationalTime(1.0, 24.0),
            duration=otio.opentime.RationalTime(24.0, 24.0)
        )
        cl = otio.schema.Clip(
            source_range=tr
        )
        cl.metadata['cmx_3600'] = {
            'reel': 'v330_21f'
        }
        tl.tracks[0].append(cl)

        result = otio.adapters.write_to_string(
            tl,
            adapter_name='cmx_3600',
            style='nucoda'
        )

        self.assertEqual(
            result,
            '001  v330_21f V     C        '
            '00:00:00:01 00:00:01:01 00:00:00:00 00:00:01:00\n'
        )

    def test_invalid_edl_style_raises_exception(self):
        tl = otio.adapters.read_from_string(
            '001  AX       V     C        '
            '00:00:00:00 00:00:00:05 00:00:00:00 00:00:00:05\n',
            adapter_name="cmx_3600"
        )
        with self.assertRaises(otio.exceptions.NotSupportedError):
            otio.adapters.write_to_string(
                tl,
                adapter_name='cmx_3600',
                style='bogus'
            )

    def test_invalid_record_timecode(self):
        with self.assertRaises(ValueError):
            tl = otio.adapters.read_from_file(TIMECODE_MISMATCH_TEST)
        with self.assertRaises(cmx_3600.EDLParseError):
            tl = otio.adapters.read_from_file(TIMECODE_MISMATCH_TEST, rate=25)
        tl = otio.adapters.read_from_file(
            TIMECODE_MISMATCH_TEST,
            rate=25,
            ignore_timecode_mismatch=True
        )
        self.assertEqual(
            tl.tracks[0][3].range_in_parent(),
            otio.opentime.TimeRange(
                start_time=otio.opentime.from_timecode("00:00:17:22", 25),
                duration=otio.opentime.from_timecode("00:00:01:24", 25)
            )
        )

    def test_can_read_frame_cut_points(self):
        # EXERCISE
        tl = otio.adapters.read_from_string(
            '1 CLPA V C     113 170 0 57\n'
            '2 CLPA V C     170 170 57 57\n'
            '2 CLPB V D 027 162 189 57 84\n'
            '3 CLPB V C     189 381 84 276\n',
            adapter_name="cmx_3600"
        )

        # VALIDATE
        self.assertEqual(tl.duration().value, 276)
        self.assertEqual(len(tl.tracks[0]), 4)
        self.assertEqual(tl.tracks[0][0].duration().value, 57)
        self.assertEqual(tl.tracks[0][0].visible_range().duration.value, 57 + 27)
        self.assertEqual(tl.tracks[0][1].in_offset.value, 0)
        self.assertEqual(tl.tracks[0][1].out_offset.value, 27)
        self.assertEqual(tl.tracks[0][2].duration().value, 27)
        self.assertEqual(tl.tracks[0][3].duration().value, 276 - 84)

    def test_speed_effects(self):
        tl = otio.adapters.read_from_file(
            SPEED_EFFECTS_TEST
        )
        self.assertEqual(
            tl.duration(),
            otio.opentime.from_timecode("00:21:03:18", 24)
        )

        # Look for a clip with a freeze frame effect
        clip = tl.tracks[0][182]
        self.assertEqual(clip.name, "Z682_156 (LAY3)")
        self.assertTrue(
            clip.effects and clip.effects[0].effect_name == 'FreezeFrame'
        )
        self.assertEqual(
            clip.duration(),
            otio.opentime.from_timecode("00:00:00:17", 24)
        )
        clip = tl.tracks[0][182]
        # TODO: We should be able to ask for the source without the effect
        # self.assertEqual(
        #     clip.source_range,
        #     otio.opentime.TimeRange(
        #         start_time=otio.opentime.from_timecode("01:00:10:21", 24),
        #         duration=otio.opentime.from_timecode("00:00:00:01", 24)
        #     )
        # )
        self.assertEqual(
            clip.range_in_parent(),
            otio.opentime.TimeRange(
                start_time=otio.opentime.from_timecode("00:08:30:00", 24),
                duration=otio.opentime.from_timecode("00:00:00:17", 24)
            )
        )

        # Look for a clip with an M2 effect
        clip = tl.tracks[0][281]
        self.assertEqual(
            clip.name,
            "Z686_5A (LAY2) (47.56 FPS)"
        )
        self.assertTrue(
            clip.effects and clip.effects[0].effect_name == "LinearTimeWarp"
        )
        self.assertAlmostEqual(clip.effects[0].time_scalar, 1.98333333)

        self.assertIsNone(
            clip.metadata.get("cmx_3600", {}).get("motion")
        )
        self.assertEqual(
            clip.duration(),
            otio.opentime.from_timecode("00:00:01:12", 24)
        )
        # TODO: We should be able to ask for the source without the effect
        # self.assertEqual(
        #     clip.source_range,
        #     otio.opentime.TimeRange(
        #         start_time=otio.opentime.from_timecode("01:00:06:00", 24),
        #         duration=otio.opentime.from_timecode("00:00:02:22", 24)
        #     )
        # )
        self.assertEqual(
            clip.range_in_parent(),
            otio.opentime.TimeRange(
                start_time=otio.opentime.from_timecode("00:11:31:16", 24),
                duration=otio.opentime.from_timecode("00:00:01:12", 24)
            )
        )

    def test_transition_duration(self):
        tl = otio.adapters.read_from_file(TRANSITION_DURATION_TEST)
        self.assertEqual(len(tl.tracks[0]), 5)

        self.assertIsInstance(tl.tracks[0][2], otio.schema.Transition)

        self.assertEqual(tl.tracks[0][2].duration().value, 26.0)

    def test_three_part_transition(self):
        """
        Test A->B->C Transition
        """
        tl = otio.adapters.read_from_file(DISSOLVE_TEST_4)
        self.assertEqual(len(tl.tracks[0]), 8)

        self.assertEqual(tl.tracks[0][0].duration().value, 30.0)
        self.assertEqual(tl.tracks[0][1].duration().value, 51.0)
        self.assertEqual(tl.tracks[0][1].visible_range().duration.value, 51 + 35)
        self.assertIsInstance(tl.tracks[0][2], otio.schema.Transition)
        self.assertEqual(tl.tracks[0][2].duration().value, 35.0)
        self.assertEqual(tl.tracks[0][3].duration().value, 81.0)
        self.assertEqual(tl.tracks[0][3].visible_range().duration.value, 81 + 64)
        self.assertIsInstance(tl.tracks[0][4], otio.schema.Transition)
        self.assertEqual(tl.tracks[0][4].duration().value, 64.0)
        self.assertEqual(tl.tracks[0][5].duration().value, 84.0)
        self.assertEqual(tl.tracks[0][5].visible_range().duration.value, 84.0)
        self.assertEqual(tl.tracks[0][6].duration().value, 96.0)
        self.assertEqual(tl.tracks[0][7].duration().value, 135.0)

    def test_enabled(self):
        tl = otio.adapters.read_from_file(ENABLED_TEST)
        # Exception is raised because the OTIO file has two tracks and cmx_3600 only
        # supports one
        with self.assertRaises(otio.exceptions.NotSupportedError):
            otio.adapters.write_to_string(tl, adapter_name="cmx_3600")

        # Disable top track so we only have one track
        tl.tracks[1].enabled = False
        result = otio.adapters.write_to_string(tl, adapter_name="cmx_3600")
        expected = r'''TITLE: enable_test

001  Clip001  V     C        00:00:00:00 00:00:00:03 00:00:00:00 00:00:00:03
* FROM CLIP NAME:  Clip-001
* OTIO TRUNCATED REEL NAME FROM: Clip-001
002  Clip002  V     C        00:00:00:03 00:00:00:06 00:00:00:03 00:00:00:06
* FROM CLIP NAME:  Clip-002
* OTIO TRUNCATED REEL NAME FROM: Clip-002
'''

        self.assertMultiLineEqual(result, expected)

        # Disable first clip in the track
        tl.tracks[0].children_if()[0].enabled = False
        result = otio.adapters.write_to_string(tl, adapter_name="cmx_3600")
        expected = r'''TITLE: enable_test

001  Clip002  V     C        00:00:00:03 00:00:00:06 00:00:00:03 00:00:00:06
* FROM CLIP NAME:  Clip-002
* OTIO TRUNCATED REEL NAME FROM: Clip-002
'''

        self.assertMultiLineEqual(result, expected)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
#
# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Test file for the track algorithms library."""

import unittest
import copy

import opentimelineio as otio
import opentimelineio.test_utils as otio_test_utils

# for debugging
# def print_expanded_tree(seq):
#     for thing in seq:
#         if isinstance(thing, tuple):
#             for i in thing:
#                 print "  ", i
#         else:
#             print thing


class TransitionExpansionTests(unittest.TestCase):
    """ test harness for transition expansion function """

    def test_expand_surrounded_by_clips(self):
        name = "test"
        rt = otio.opentime.RationalTime(5, 24)
        rt_2 = otio.opentime.RationalTime(1, 24)
        tr = otio.opentime.TimeRange(rt, rt + rt)
        avail_tr = otio.opentime.TimeRange(
            otio.opentime.RationalTime(0, 24),
            otio.opentime.RationalTime(50, 24)
        )
        mr = otio.schema.ExternalReference(
            available_range=avail_tr,
            target_url="/var/tmp/test.mov"
        )

        cl = otio.schema.Clip(
            name=name + "_pre",
            media_reference=mr,
            source_range=tr,
        )

        seq = otio.schema.Track()
        seq.append(cl)
        in_offset = rt
        out_offset = rt_2
        trx = otio.schema.Transition(
            name="AtoB",
            transition_type=otio.schema.TransitionTypes.SMPTE_Dissolve,
            in_offset=in_offset,
            out_offset=out_offset,
            metadata={
                "foo": "bar"
            }
        )
        seq.append(trx)
        cl_2 = copy.deepcopy(cl)
        cl_2.name = name + "_post"
        seq.append(copy.deepcopy(cl))

        pre_duration = copy.deepcopy(seq[0].source_range.duration)

        # print
        # print "BEFORE:"
        # print_expanded_tree(seq)
        seq2 = otio.algorithms.track_with_expanded_transitions(seq)
        # print
        # print "AFTER:"
        # print_expanded_tree(seq2)

        self.assertNotEqual(pre_duration, seq2[0].source_range.duration)

        # check ranges on results

        # first clip is trimmed
        self.assertEqual(
            seq2[0].source_range.duration,
            cl_2.source_range.duration - in_offset
        )

        self.assertEqual(
            seq2[1][0].source_range.start_time,
            cl.source_range.end_time_exclusive() - in_offset
        )
        self.assertEqual(
            seq2[1][0].source_range.duration,
            in_offset + out_offset
        )
        self.assertEqual(
            seq2[1][2].source_range.start_time,
            cl_2.source_range.start_time - in_offset
        )
        self.assertEqual(
            seq2[1][2].source_range.duration,
            in_offset + out_offset
        )

        # final clip is trimmed
        self.assertEqual(
            seq2[2].source_range.duration,
            cl_2.source_range.duration - out_offset
        )

        # make sure that both transition clips are the same length
        self.assertEqual(
            seq2[1][0].source_range.duration,
            seq2[1][2].source_range.duration
        )

        # big hammer stuff
        self.assertNotEqual(seq, seq2)
        self.assertNotEqual(seq[0], seq2[0])
        self.assertNotEqual(seq[-1], seq2[-1])
        self.assertNotEqual(seq[-1].source_range, seq2[-1].source_range)

    # leaving this here as a @TODO
    def DISABLED_test_expand_track(self):
        name = "test"
        rt = otio.opentime.RationalTime(5, 24)
        rt_2 = otio.opentime.RationalTime(1, 24)
        tr = otio.opentime.TimeRange(rt, rt + rt)
        avail_tr = otio.opentime.TimeRange(
            otio.opentime.RationalTime(0, 24),
            otio.opentime.RationalTime(50, 24)
        )
        mr = otio.schema.ExternalReference(
            available_range=avail_tr,
            target_url="/var/tmp/test.mov"
        )

        cl = otio.schema.Clip(
            name=name + "_pre",
            media_reference=mr,
            source_range=tr,
        )

        seq = otio.schema.Track()
        seq.name = name
        seq.append(cl)
        in_offset = rt
        out_offset = rt_2
        trx = otio.schema.Transition(
            name="AtoB",
            transition_type=otio.schema.TransitionTypes.SMPTE_Dissolve,
            in_offset=in_offset,
            out_offset=out_offset,
            metadata={
                "foo": "bar"
            }
        )
        seq.append(trx)
        seq_2 = copy.deepcopy(seq)
        seq_2.name = name + "_post"
        seq.append(seq_2)

        pre_duration = copy.deepcopy(seq[0].source_range.duration)

        # print
        # print "BEFORE:"
        # print_expanded_tree(seq)
        expanded_seq = otio.algorithms.track_with_expanded_transitions(seq)
        # print
        # print "AFTER:"
        # print_expanded_tree(expanded_seq)

        self.assertNotEqual(
            pre_duration,
            expanded_seq[0].source_range.duration
        )

        # check ranges on results

        # first clip is trimmed
        self.assertEqual(
            expanded_seq[0].source_range.duration,
            seq_2.source_range.duration - in_offset
        )

        self.assertEqual(
            expanded_seq[1][0].source_range.start_time,
            cl.source_range.end_time_exclusive() - in_offset
        )
        self.assertEqual(
            expanded_seq[1][0].source_range.duration,
            in_offset + out_offset
        )
        self.assertEqual(
            expanded_seq[1][2].source_range.start_time,
            seq_2.source_range.start_time - in_offset
        )
        self.assertEqual(
            expanded_seq[1][2].source_range.duration,
            in_offset + out_offset
        )

        # final clip is trimmed
        self.assertEqual(
            expanded_seq[2].source_range.duration,
            seq_2.source_range.duration - out_offset
        )

        # make sure that both transition clips are the same length
        self.assertEqual(
            expanded_seq[1][0].source_range.duration,
            expanded_seq[1][2].source_range.duration
        )

        # big hammer stuff
        self.assertNotEqual(seq, expanded_seq)
        self.assertNotEqual(seq[0], expanded_seq[0])
        self.assertNotEqual(seq[-1], expanded_seq[-1])
        self.assertNotEqual(
            seq[-1].source_range,
            expanded_seq[-1].source_range
        )


class TrackTrimmingTests(unittest.TestCase, otio_test_utils.OTIOAssertions):
    """ test harness for track trimming function """

    def make_sample_track(self):
        return otio.adapters.read_from_string("""
        {
            "OTIO_SCHEMA": "Track.1",
            "children": [
                {
                    "OTIO_SCHEMA": "Clip.1",
                    "effects": [],
                    "markers": [],
                    "media_reference": null,
                    "metadata": {},
                    "name": "A",
                    "source_range": {
                        "OTIO_SCHEMA": "TimeRange.1",
                        "duration": {
                            "OTIO_SCHEMA": "RationalTime.1",
                            "rate": 24,
                            "value": 50
                        },
                        "start_time": {
                            "OTIO_SCHEMA": "RationalTime.1",
                            "rate": 24,
                            "value": 0.0
                        }
                    }
                },
                {
                    "OTIO_SCHEMA": "Clip.1",
                    "effects": [],
                    "markers": [],
                    "media_reference": null,
                    "metadata": {},
                    "name": "B",
                    "source_range": {
                        "OTIO_SCHEMA": "TimeRange.1",
                        "duration": {
                            "OTIO_SCHEMA": "RationalTime.1",
                            "rate": 24,
                            "value": 50
                        },
                        "start_time": {
                            "OTIO_SCHEMA": "RationalTime.1",
                            "rate": 24,
                            "value": 0.0
                        }
                    }
                },
                {
                    "OTIO_SCHEMA": "Clip.1",
                    "effects": [],
                    "markers": [],
                    "media_reference": null,
                    "metadata": {},
                    "name": "C",
                    "source_range": {
                        "OTIO_SCHEMA": "TimeRange.1",
                        "duration": {
                            "OTIO_SCHEMA": "RationalTime.1",
                            "rate": 24,
                            "value": 50
                        },
                        "start_time": {
                            "OTIO_SCHEMA": "RationalTime.1",
                            "rate": 24,
                            "value": 0.0
                        }
                    }
                }
            ],
            "effects": [],
            "kind": "Video",
            "markers": [],
            "metadata": {},
            "name": "Sequence1",
            "source_range": null
        }
        """, "otio_json")

    def test_trim_to_existing_range(self):
        original_track = self.make_sample_track()
        self.assertEqual(
            original_track.trimmed_range(),
            otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(0, 24),
                duration=otio.opentime.RationalTime(150, 24)
            )
        )

        # trim to the exact range it already has
        trimmed = otio.algorithms.track_trimmed_to_range(
            original_track,
            otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(0, 24),
                duration=otio.opentime.RationalTime(150, 24)
            )
        )
        # it shouldn't have changed at all
        self.assertIsOTIOEquivalentTo(original_track, trimmed)

    def test_trim_to_longer_range(self):
        original_track = self.make_sample_track()
        # trim to a larger range
        trimmed = otio.algorithms.track_trimmed_to_range(
            original_track,
            otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(-10, 24),
                duration=otio.opentime.RationalTime(160, 24)
            )
        )
        # it shouldn't have changed at all
        self.assertJsonEqual(original_track, trimmed)

    def test_trim_front(self):
        original_track = self.make_sample_track()
        # trim off the front (clip A and part of B)
        trimmed = otio.algorithms.track_trimmed_to_range(
            original_track,
            otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(60, 24),
                duration=otio.opentime.RationalTime(90, 24)
            )
        )
        self.assertNotEqual(original_track, trimmed)
        self.assertEqual(len(trimmed), 2)
        self.assertEqual(
            trimmed.trimmed_range(),
            otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(0, 24),
                duration=otio.opentime.RationalTime(90, 24)
            )
        )
        # did clip B get trimmed?
        self.assertEqual(trimmed[0].name, "B")
        self.assertEqual(
            trimmed[0].trimmed_range(),
            otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(10, 24),
                duration=otio.opentime.RationalTime(40, 24)
            )
        )
        # clip C should have been left alone
        self.assertIsOTIOEquivalentTo(trimmed[1], original_track[2])

    def test_trim_end(self):
        original_track = self.make_sample_track()
        # trim off the end (clip C and part of B)
        trimmed = otio.algorithms.track_trimmed_to_range(
            original_track,
            otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(0, 24),
                duration=otio.opentime.RationalTime(90, 24)
            )
        )
        self.assertNotEqual(original_track, trimmed)
        self.assertEqual(len(trimmed), 2)
        self.assertEqual(
            trimmed.trimmed_range(),
            otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(0, 24),
                duration=otio.opentime.RationalTime(90, 24)
            )
        )
        # clip A should have been left alone
        self.assertIsOTIOEquivalentTo(trimmed[0], original_track[0])
        # did clip B get trimmed?
        self.assertEqual(trimmed[1].name, "B")
        self.assertEqual(
            trimmed[1].trimmed_range(),
            otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(0, 24),
                duration=otio.opentime.RationalTime(40, 24)
            )
        )

    def test_trim_with_transitions(self):
        original_track = self.make_sample_track()
        self.assertEqual(
            otio.opentime.RationalTime(150, 24),
            original_track.duration()
        )
        self.assertEqual(len(original_track), 3)

        # add a transition
        tr = otio.schema.Transition(
            in_offset=otio.opentime.RationalTime(12, 24),
            out_offset=otio.opentime.RationalTime(20, 24)
        )
        original_track.insert(1, tr)
        self.assertEqual(len(original_track), 4)
        self.assertEqual(
            otio.opentime.RationalTime(150, 24),
            original_track.duration()
        )

        # if you try to sever a Transition in the middle it should fail
        with self.assertRaises(otio.exceptions.CannotTrimTransitionsError):
            trimmed = otio.algorithms.track_trimmed_to_range(
                original_track,
                otio.opentime.TimeRange(
                    start_time=otio.opentime.RationalTime(5, 24),
                    duration=otio.opentime.RationalTime(50, 24)
                )
            )

        with self.assertRaises(otio.exceptions.CannotTrimTransitionsError):
            trimmed = otio.algorithms.track_trimmed_to_range(
                original_track,
                otio.opentime.TimeRange(
                    start_time=otio.opentime.RationalTime(45, 24),
                    duration=otio.opentime.RationalTime(50, 24)
                )
            )

        trimmed = otio.algorithms.track_trimmed_to_range(
            original_track,
            otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(25, 24),
                duration=otio.opentime.RationalTime(50, 24)
            )
        )
        self.assertNotEqual(original_track, trimmed)

        expected = otio.adapters.read_from_string("""
        {
            "OTIO_SCHEMA": "Track.1",
            "children": [
                {
                    "OTIO_SCHEMA": "Clip.1",
                    "effects": [],
                    "markers": [],
                    "media_reference": null,
                    "metadata": {},
                    "name": "A",
                    "source_range": {
                        "OTIO_SCHEMA": "TimeRange.1",
                        "duration": {
                            "OTIO_SCHEMA": "RationalTime.1",
                            "rate": 24,
                            "value": 25
                        },
                        "start_time": {
                            "OTIO_SCHEMA": "RationalTime.1",
                            "rate": 24,
                            "value": 25.0
                        }
                    }
                },
                {
                    "OTIO_SCHEMA": "Transition.1",
                    "name": "",
                    "metadata": {},
                    "transition_type": "",
                    "in_offset": {
                        "OTIO_SCHEMA": "RationalTime.1",
                        "rate": 24,
                        "value": 12
                    },
                    "out_offset": {
                        "OTIO_SCHEMA": "RationalTime.1",
                        "rate": 24,
                        "value": 20
                    }
                },
                {
                    "OTIO_SCHEMA": "Clip.1",
                    "effects": [],
                    "markers": [],
                    "media_reference": null,
                    "metadata": {},
                    "name": "B",
                    "source_range": {
                        "OTIO_SCHEMA": "TimeRange.1",
                        "duration": {
                            "OTIO_SCHEMA": "RationalTime.1",
                            "rate": 24,
                            "value": 25
                        },
                        "start_time": {
                            "OTIO_SCHEMA": "RationalTime.1",
                            "rate": 24,
                            "value": 0.0
                        }
                    }
                }
            ],
            "effects": [],
            "kind": "Video",
            "markers": [],
            "metadata": {},
            "name": "Sequence1",
            "source_range": null
        }
        """, "otio_json")

        self.assertJsonEqual(expected, trimmed)


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

""" Generator Reference class test harness.  """

import unittest
import os

import opentimelineio as otio
import opentimelineio.test_utils as otio_test_utils

SAMPLE_DATA_DIR = os.path.join(os.path.dirname(__file__), "sample_data")
GEN_REF_TEST = os.path.join(SAMPLE_DATA_DIR, "generator_reference_test.otio")


class GeneratorRefTests(unittest.TestCase, otio_test_utils.OTIOAssertions):
    def setUp(self):
        self.gen = otio.schema.GeneratorReference(
            name="SMPTEBars",
            generator_kind="SMPTEBars",
            available_range=otio.opentime.TimeRange(
                otio.opentime.RationalTime(0, 24),
                otio.opentime.RationalTime(100, 24),
            ),
            parameters={
                "test_param": 5.0,
            },
            metadata={
                "foo": "bar"
            },
            available_image_bounds=otio.schema.Box2d(
                otio.schema.V2d(0.0, 0.0),
                otio.schema.V2d(16.0, 9.0)
            )
        )

    def test_constructor(self):
        self.assertEqual(self.gen.generator_kind, "SMPTEBars")
        self.assertEqual(self.gen.name, "SMPTEBars")
        self.assertEqual(self.gen.parameters, {"test_param": 5.0})
        self.assertEqual(self.gen.metadata, {"foo": "bar"})
        self.assertEqual(
            self.gen.available_range,
            otio.opentime.TimeRange(
                otio.opentime.RationalTime(0, 24),
                otio.opentime.RationalTime(100, 24),
            )
        )
        self.assertEqual(
            self.gen.available_image_bounds,
            otio.schema.Box2d(
                otio.schema.V2d(0.0, 0.0),
                otio.schema.V2d(16.0, 9.0)
            )
        )

    def test_serialize(self):
        encoded = otio.adapters.otio_json.write_to_string(self.gen)
        decoded = otio.adapters.otio_json.read_from_string(encoded)
        self.assertIsOTIOEquivalentTo(self.gen, decoded)

    def test_read_file(self):
        self.assertTrue(os.path.exists(GEN_REF_TEST))
        decoded = otio.adapters.otio_json.read_from_file(GEN_REF_TEST)
        self.assertEqual(
            decoded.tracks[0][0].media_reference.generator_kind, "SMPTEBars"
        )

    def test_stringify(self):
        self.assertMultiLineEqual(
            str(self.gen),
            "GeneratorReference("
            '"{}", '
            '"{}", '
            '{}, '
            '{}, '
            "{}"
            ")".format(
                str(self.gen.name),
                str(self.gen.generator_kind),
                str(self.gen.parameters),
                str(self.gen.available_image_bounds),
                str(self.gen.metadata),
            )
        )

        self.assertMultiLineEqual(
            repr(self.gen),
            "otio.schema.GeneratorReference("
            "name={}, "
            "generator_kind={}, "
            "parameters={}, "
            "available_image_bounds={}, "
            "metadata={}"
            ")".format(
                repr(self.gen.name),
                repr(self.gen.generator_kind),
                repr(self.gen.parameters),
                repr(self.gen.available_image_bounds),
                repr(self.gen.metadata),
            )
        )


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Test the ALE adapter."""

# python
import os
import unittest

import opentimelineio as otio

SAMPLE_DATA_DIR = os.path.join(os.path.dirname(__file__), "sample_data")
EXAMPLE_PATH = os.path.join(SAMPLE_DATA_DIR, "sample.ale")
EXAMPLE2_PATH = os.path.join(SAMPLE_DATA_DIR, "sample2.ale")
EXAMPLE_CDL_PATH = os.path.join(SAMPLE_DATA_DIR, "sample_cdl.ale")
EXAMPLEUHD_PATH = os.path.join(SAMPLE_DATA_DIR, "sampleUHD.ale")


class ALEAdapterTest(unittest.TestCase):

    def test_ale_read(self):
        ale_path = EXAMPLE_PATH
        collection = otio.adapters.read_from_file(ale_path)
        self.assertTrue(collection is not None)
        self.assertEqual(type(collection), otio.schema.SerializableCollection)
        self.assertEqual(len(collection), 4)
        fps = float(collection.metadata.get("ALE").get("header").get("FPS"))
        self.assertEqual(fps, 24)
        self.assertEqual(
            [c.name for c in collection],
            ["test_017056", "test_017057", "test_017058", "Something"]
        )
        self.assertEqual(
            [c.source_range for c in collection],
            [
                otio.opentime.TimeRange(
                    otio.opentime.from_timecode("01:00:00:00", fps),
                    otio.opentime.from_timecode("00:00:04:03", fps)
                ),
                otio.opentime.TimeRange(
                    otio.opentime.from_timecode("01:00:00:00", fps),
                    otio.opentime.from_timecode("00:00:04:04", fps)
                ),
                otio.opentime.TimeRange(
                    otio.opentime.from_timecode("01:00:00:00", fps),
                    otio.opentime.from_timecode("00:00:04:05", fps)
                ),
                otio.opentime.TimeRange(
                    otio.opentime.from_timecode("01:00:00:00", fps),
                    otio.opentime.from_timecode("00:00:04:06", fps)
                )
            ]
        )

    def test_ale_read2(self):
        ale_path = EXAMPLE2_PATH
        collection = otio.adapters.read_from_file(ale_path)
        self.assertTrue(collection is not None)
        self.assertEqual(type(collection), otio.schema.SerializableCollection)
        self.assertEqual(len(collection), 2)
        fps = float(collection.metadata.get("ALE").get("header").get("FPS"))
        self.assertEqual(fps, 23.98)
        self.assertEqual(
            [c.name for c in collection],
            ["19A-1xa", "19A-2xa"]
        )
        self.assertEqual(
            [c.source_range for c in collection],
            [
                otio.opentime.TimeRange(
                    otio.opentime.from_timecode("04:00:00:00", fps),
                    otio.opentime.from_timecode("00:00:46:16", fps)
                ),
                otio.opentime.TimeRange(
                    otio.opentime.from_timecode("04:00:46:16", fps),
                    otio.opentime.from_timecode("00:00:50:16", fps)
                )
            ]
        )

    def test_ale_read_cdl(self):
        ale_path = EXAMPLE_CDL_PATH
        collection = otio.adapters.read_from_file(ale_path)
        self.assertTrue(collection is not None)
        self.assertEqual(type(collection), otio.schema.SerializableCollection)
        self.assertEqual(len(collection), 4)
        fps = float(collection.metadata.get("ALE").get("header").get("FPS"))
        self.assertEqual(fps, 23.976)
        self.assertEqual([c.name for c in collection], [
            "A005_C010_0501J0", "A005_C010_0501J0", "A005_C009_0501A0",
            "A005_C010_0501J0"
        ])
        self.assertEqual([c.source_range for c in collection], [

            otio.opentime.TimeRange(
                otio.opentime.from_timecode("17:49:33:01", fps),
                otio.opentime.from_timecode("00:00:02:09", fps)),

            otio.opentime.TimeRange(
                otio.opentime.from_timecode("17:49:55:19", fps),
                otio.opentime.from_timecode("00:00:06:09", fps)),

            otio.opentime.TimeRange(
                otio.opentime.from_timecode("17:40:25:06", fps),
                otio.opentime.from_timecode("00:00:02:20", fps)),

            otio.opentime.TimeRange(
                otio.opentime.from_timecode("17:50:21:23", fps),
                otio.opentime.from_timecode("00:00:03:14", fps))
        ])

        # Slope, offset, and power values are of type _otio.AnyVector
        # So we have to convert them to lists otherwise
        # the comparison between those two types would fail

        # FIRST CLIP
        self.assertEqual(
            list(collection[0].metadata['cdl']['asc_sop']['slope']),
            [0.8714, 0.9334, 0.9947])
        self.assertEqual(
            list(collection[0].metadata['cdl']['asc_sop']['offset']),
            [-0.087, -0.0922, -0.0808])
        self.assertEqual(
            list(collection[0].metadata['cdl']['asc_sop']['power']),
            [0.9988, 1.0218, 1.0101])
        self.assertEqual(collection[0].metadata['cdl']['asc_sat'], 0.9)

        # SECOND CLIP
        self.assertEqual(
            list(collection[1].metadata['cdl']['asc_sop']['slope']),
            [0.8714, 0.9334, 0.9947])
        self.assertEqual(
            list(collection[1].metadata['cdl']['asc_sop']['offset']),
            [-0.087, -0.0922, -0.0808])
        self.assertEqual(
            list(collection[1].metadata['cdl']['asc_sop']['power']),
            [0.9988, 1.0218, 1.0101])
        self.assertEqual(collection[1].metadata['cdl']['asc_sat'], 0.9)

        # THIRD CLIP
        self.assertEqual(
            list(collection[2].metadata['cdl']['asc_sop']['slope']),
            [0.8604, 0.9252, 0.9755])
        self.assertEqual(
            list(collection[2].metadata['cdl']['asc_sop']['offset']),
            [-0.0735, -0.0813, -0.0737])
        self.assertEqual(
            list(collection[2].metadata['cdl']['asc_sop']['power']),
            [0.9988, 1.0218, 1.0101])
        self.assertEqual(collection[2].metadata['cdl']['asc_sat'], 0.9)

        # FOURTH CLIP
        self.assertEqual(
            list(collection[3].metadata['cdl']['asc_sop']['slope']),
            [0.8714, 0.9334, 0.9947])
        self.assertEqual(
            list(collection[3].metadata['cdl']['asc_sop']['offset']),
            [-0.087, -0.0922, -0.0808])
        self.assertEqual(
            list(collection[3].metadata['cdl']['asc_sop']['power']),
            [0.9988, 1.0218, 1.0101])
        self.assertEqual(collection[3].metadata['cdl']['asc_sat'], 0.9)

    def test_ale_uhd(self):
        ale_path = EXAMPLEUHD_PATH
        collection = otio.adapters.read_from_file(ale_path)
        frmt = str(collection.metadata.get("ALE").get("header").get("VIDEO_FORMAT"))
        self.assertEqual(frmt, "CUSTOM")

    def test_ale_add_format(self):

        # adds a clip to the supplied timeline, sets the clips "Image Size"
        # metadata and then rountrips the ALE verifying the supplied format is detected
        def add_then_check(timeline, size, expected_format):
            cl = otio.schema.Clip(
                metadata={'ALE': {'Image Size': size}},
                source_range=otio.opentime.TimeRange(
                    start_time=otio.opentime.RationalTime(0, 23.976),
                    duration=otio.opentime.RationalTime(48, 23.976)
                )
            )
            timeline.tracks[0].extend([cl])
            collection = otio.adapters.read_from_string(
                otio.adapters.write_to_string(
                    timeline,
                    adapter_name='ale'
                ),
                adapter_name="ale"
            )
            ale_meta = collection.metadata.get('ALE')
            vid_format = str(ale_meta.get('header').get('VIDEO_FORMAT'))
            self.assertEqual(vid_format, expected_format)

        track = otio.schema.Track()
        tl = otio.schema.Timeline("Add Format", tracks=[track])

        # add multiple clips with various resolutions,
        # we want the ALE to return a project format
        # that is compatible with the largest resolution

        add_then_check(tl, '720 x 486', 'NTSC')
        add_then_check(tl, '720 x 576', 'PAL')
        add_then_check(tl, '1280x 720', '720')
        add_then_check(tl, '1920x1080', '1080')
        add_then_check(tl, '2048x1080', 'CUSTOM')
        add_then_check(tl, '4096x2304', 'CUSTOM')

    def test_ale_roundtrip(self):
        ale_path = EXAMPLE_PATH

        with open(ale_path, 'r') as fi:
            original = fi.read()
            collection = otio.adapters.read_from_string(original, "ale")
            output = otio.adapters.write_to_string(collection, "ale")
            self.maxDiff = None
            self.assertMultiLineEqual(original, output)


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Test cases to verify examples used in the OTIO documentation."""

import os
import unittest

import opentimelineio as otio


SAMPLE_DATA_DIR = os.path.join(os.path.dirname(__file__), "sample_data")
CLIP_EXAMPLE_PATH = os.path.join(SAMPLE_DATA_DIR, "clip_example.otio")


class DocTester(unittest.TestCase):

    def test_clip(self):
        timeline = otio.adapters.read_from_file(CLIP_EXAMPLE_PATH)
        track = timeline.tracks[0]
        gapA, clip, transition, gapB = track[:]

        self.assertEqual(
            otio.opentime.RationalTime(19, 24),
            track.duration()
        )
        self.assertEqual(
            otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(0, 24),
                duration=otio.opentime.RationalTime(19, 24)
            ),
            track.trimmed_range()
        )
        self.assertEqual(
            otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(0, 24),
                duration=otio.opentime.RationalTime(19, 24)
            ),
            track.available_range()
        )
        self.assertEqual(
            otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(0, 24),
                duration=otio.opentime.RationalTime(19, 24)
            ),
            track.visible_range()
        )

        self.assertEqual(
            otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(8, 24),
                duration=otio.opentime.RationalTime(3, 24)
            ),
            track.trimmed_range_of_child(clip)
        )
        self.assertEqual(
            (
                None,
                otio.opentime.RationalTime(1, 24)
            ),
            track.handles_of_child(clip)
        )
        self.assertEqual(
            otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(8, 24),
                duration=otio.opentime.RationalTime(3, 24)
            ),
            track.trimmed_range_of_child(clip)
        )
        self.assertEqual(
            otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(8, 24),
                duration=otio.opentime.RationalTime(3, 24)
            ),
            track.range_of_child(clip)
        )

        self.assertEqual(
            otio.opentime.RationalTime(8, 24),
            gapA.duration()
        )

        self.assertEqual(
            otio.opentime.RationalTime(3, 24),
            clip.duration()
        )
        self.assertEqual(
            otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(8, 24),
                duration=otio.opentime.RationalTime(3, 24)
            ),
            clip.trimmed_range_in_parent()
        )
        self.assertEqual(
            otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(8, 24),
                duration=otio.opentime.RationalTime(3, 24)
            ),
            clip.range_in_parent()
        )
        self.assertEqual(
            otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(3, 24),
                duration=otio.opentime.RationalTime(3, 24)
            ),
            clip.trimmed_range()
        )
        self.assertEqual(
            otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(3, 24),
                duration=otio.opentime.RationalTime(4, 24)
            ),
            clip.visible_range()
        )
        self.assertEqual(
            otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(0, 24),
                duration=otio.opentime.RationalTime(8, 24)
            ),
            clip.available_range()
        )
        self.assertEqual(
            otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(0, 24),
                duration=otio.opentime.RationalTime(8, 24)
            ),
            clip.media_reference.available_range
        )

        self.assertEqual(
            otio.opentime.RationalTime(2, 24),
            transition.in_offset
        )
        self.assertEqual(
            otio.opentime.RationalTime(1, 24),
            transition.out_offset
        )

        self.assertEqual(
            otio.opentime.RationalTime(8, 24),
            gapB.duration()
        )


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

import os

"""This file is here to support the test_adapter_plugin unittest.
If you want to learn how to write your own adapter plugin, please read:
https://opentimelineio.readthedocs.io/en/latest/tutorials/write-an-adapter.html
"""


def hook_function(in_timeline, argument_map=None):
    filepath = argument_map.get('_filepath')
    argument_map.update({'filesize': os.path.getsize(filepath)})
    in_timeline.metadata.update(argument_map)

    return in_timeline



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Test harness for Media References."""

import opentimelineio as otio
import opentimelineio.test_utils as otio_test_utils

import unittest


class MediaReferenceTests(unittest.TestCase, otio_test_utils.OTIOAssertions):

    def test_cons(self):
        tr = otio.opentime.TimeRange(
            otio.opentime.RationalTime(5, 24),
            otio.opentime.RationalTime(10, 24.0)
        )
        mr = otio.schema.MissingReference(
            available_range=tr,
            metadata={'show': 'OTIOTheMovie'}
        )

        self.assertEqual(mr.available_range, tr)

        mr = otio.schema.MissingReference()
        self.assertIsNone(mr.available_range)

    def test_str_missing(self):
        missing = otio.schema.MissingReference()
        self.assertMultiLineEqual(
            str(missing),
            "MissingReference(\'\', None, None, {})"
        )
        self.assertMultiLineEqual(
            repr(missing),
            "otio.schema.MissingReference("
            "name='', available_range=None, available_image_bounds=None, metadata={}"
            ")"
        )

        encoded = otio.adapters.otio_json.write_to_string(missing)
        decoded = otio.adapters.otio_json.read_from_string(encoded)
        self.assertIsOTIOEquivalentTo(missing, decoded)

    def test_filepath(self):
        filepath = otio.schema.ExternalReference("/var/tmp/foo.mov")
        self.assertMultiLineEqual(
            str(filepath),
            'ExternalReference("/var/tmp/foo.mov")'
        )
        self.assertMultiLineEqual(
            repr(filepath),
            "otio.schema.ExternalReference("
            "target_url='/var/tmp/foo.mov'"
            ")"
        )

        # round trip serialize
        encoded = otio.adapters.otio_json.write_to_string(filepath)
        decoded = otio.adapters.otio_json.read_from_string(encoded)
        self.assertIsOTIOEquivalentTo(filepath, decoded)

    def test_equality(self):
        filepath = otio.schema.ExternalReference(target_url="/var/tmp/foo.mov")
        filepath2 = otio.schema.ExternalReference(
            target_url="/var/tmp/foo.mov"
        )
        self.assertIsOTIOEquivalentTo(filepath, filepath2)

        bl = otio.schema.MissingReference()
        self.assertNotEqual(filepath, bl)

        filepath2 = otio.schema.ExternalReference(
            target_url="/var/tmp/foo2.mov"
        )
        self.assertNotEqual(filepath, filepath2)
        self.assertEqual(filepath == filepath2, False)

    def test_is_missing(self):
        mr = otio.schema.ExternalReference(target_url="/var/tmp/foo.mov")
        self.assertFalse(mr.is_missing_reference)

        mr = otio.schema.MissingReference()
        self.assertTrue(mr.is_missing_reference)


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

# python
import os
import unittest

import opentimelineio as otio

__doc__ = """Test CDL support in the EDL adapter."""

SAMPLE_DATA_DIR = os.path.join(os.path.dirname(__file__), "sample_data")
CDL_EXAMPLE_PATH = os.path.join(SAMPLE_DATA_DIR, "cdl.edl")


class CDLAdapterTest(unittest.TestCase):
    def test_cdl_read(self):
        edl_path = CDL_EXAMPLE_PATH
        timeline = otio.adapters.read_from_file(edl_path)
        self.assertTrue(timeline is not None)
        self.assertEqual(len(timeline.tracks), 1)
        self.assertEqual(len(timeline.tracks[0]), 2)
        for clip in timeline.tracks[0]:
            # clip = timeline.tracks[0][0]
            self.assertEqual(
                clip.name,
                "ZZ100_501 (LAY3)"
            )
            self.assertEqual(
                clip.source_range.duration,
                otio.opentime.from_timecode("00:00:01:07", 24)
            )
            cdl = clip.metadata.get("cdl", {})
            self.assertEqual(
                cdl.get("asc_sat"),
                0.9
            )
            self.assertEqual(
                list(cdl.get("asc_sop").get("slope")),
                [0.1, 0.2, 0.3]
            )
            self.assertEqual(
                list(cdl.get("asc_sop").get("offset")),
                [1.0000, -0.0122, 0.0305]
            )
            self.assertEqual(
                list(cdl.get("asc_sop").get("power")),
                [1.0000, 0.0000, 1.0000]
            )

    def test_cdl_read_with_commas(self):
        # This EDL was generated with Premiere Pro using the CDL master effect
        # on a clip
        cdl = """TITLE: Sequence 01
FCM: NON-DROP FRAME

000001  A006C014_1701069O V     C        04:34:41:13 04:34:41:16 00:00:00:00 00:00:00:03
* FROM CLIP NAME: A006C014_1701069O_LOG_NO_LUT.mov
* ASC_SOP: (1.1549, 1.1469, 1.1422000000000001)(-0.067799999999999999, -0.055500000000000001, -0.032300000000000002)(1.1325000000000001, 1.1351, 1.1221000000000001)
* ASC_SAT: 1.2988
"""  # noqa: E501
        timeline = otio.adapters.read_from_string(cdl, "cmx_3600")

        clip = timeline.tracks[0][0]
        cdl_metadata = clip.metadata["cdl"]

        ref_sop_values = {
            "slope": [
                1.1549,
                1.1469,
                1.1422000000000001,
            ],
            "offset": [
                -0.067799999999999999,
                -0.055500000000000001,
                -0.032300000000000002,
            ],
            "power": [
                1.1325000000000001,
                1.1351,
                1.1221000000000001,
            ],
        }

        self.assertAlmostEqual(cdl_metadata["asc_sat"], 1.2988)
        for function in ("slope", "offset", "power"):
            comparisons = zip(
                cdl_metadata["asc_sop"][function], ref_sop_values[function]
            )
            for value_comp, ref_comp in comparisons:
                self.assertAlmostEqual(
                    value_comp, ref_comp, msg="mismatch in {}".format(function)
                )

    def test_cdl_round_trip(self):
        original = """TITLE: Example_Screening.01

001  AX       V     C        01:00:04:05 01:00:05:12 00:00:00:00 00:00:01:07
* FROM CLIP NAME:  ZZ100_501 (LAY3)
*ASC_SOP (0.1 0.2 0.3) (1.0 -0.0122 0.0305) (1.0 0.0 1.0)
*ASC_SAT 0.9
* SOURCE FILE: ZZ100_501.LAY3.01
"""
        expected = """TITLE: Example_Screening.01

001  ZZ100501 V     C        01:00:04:05 01:00:05:12 00:00:00:00 00:00:01:07
* FROM CLIP NAME:  ZZ100_501 (LAY3)
* OTIO TRUNCATED REEL NAME FROM: ZZ100_501 (LAY3)
*ASC_SOP (0.1 0.2 0.3) (1.0 -0.0122 0.0305) (1.0 0.0 1.0)
*ASC_SAT 0.9
* SOURCE FILE: ZZ100_501.LAY3.01
"""
        timeline = otio.adapters.read_from_string(original, "cmx_3600")
        output = otio.adapters.write_to_string(timeline, "cmx_3600")
        self.assertMultiLineEqual(expected, output)


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""unit tests for the version manifest plugin system"""

import unittest
import os
import json

import opentimelineio as otio
from tests import utils


FIRST_MANIFEST = """{
    "OTIO_SCHEMA" : "PluginManifest.1",
    "version_manifests": {
        "UNIQUE_FAMILY": {
            "TEST_LABEL": {
                "second_thing": 3
            }
        },
        "LAYERED_FAMILY": {
            "June2022": {
                "SimpleClass": 2
            },
            "May2022": {
                "SimpleClass": 1
            }
        }
    }
}
"""

SECOND_MANIFEST = """{
    "OTIO_SCHEMA" : "PluginManifest.1",
    "version_manifests": {
        "LAYERED_FAMILY": {
            "May2022": {
                "SimpleClass": 2
            },
            "April2022": {
                "SimpleClass": 1
            }
        }
    }
}
"""


class TestPlugin_VersionManifest(unittest.TestCase):
    def setUp(self):
        self.bak = otio.plugins.ActiveManifest()
        self.man = utils.create_manifest()
        otio.plugins.manifest._MANIFEST = self.man

        if "OTIO_DEFAULT_TARGET_VERSION_FAMILY_LABEL" in os.environ:
            del os.environ["OTIO_DEFAULT_TARGET_VERSION_FAMILY_LABEL"]

    def tearDown(self):
        otio.plugins.manifest._MANIFEST = self.bak
        utils.remove_manifest(self.man)
        if "OTIO_DEFAULT_TARGET_VERSION_FAMILY_LABEL" in os.environ:
            del os.environ["OTIO_DEFAULT_TARGET_VERSION_FAMILY_LABEL"]

    def test_read_in_manifest(self):
        self.assertIn("TEST_FAMILY_NAME", self.man.version_manifests)
        self.assertIn(
            "TEST_LABEL",
            self.man.version_manifests["TEST_FAMILY_NAME"]
        )

    def test_full_map(self):
        d = otio.versioning.full_map()
        self.assertIn("TEST_FAMILY_NAME", d)
        self.assertIn(
            "TEST_LABEL",
            d["TEST_FAMILY_NAME"]
        )

    def test_fetch_map(self):
        self.assertEqual(
            otio.versioning.fetch_map("TEST_FAMILY_NAME", "TEST_LABEL"),
            {"ExampleSchema": 2, "EnvVarTestSchema": 1, "Clip": 1}
        )

    def test_env_variable_downgrade(self):
        @otio.core.register_type
        class EnvVarTestSchema(otio.core.SerializableObject):
            _serializable_label = "EnvVarTestSchema.2"
            foo_two = otio.core.serializable_field("foo_2")

        @otio.core.downgrade_function_from(EnvVarTestSchema, 2)
        def downgrade_2_to_1(_data_dict):
            return {"foo": _data_dict["foo_2"]}

        evt = EnvVarTestSchema()
        evt.foo_two = "asdf"

        result = json.loads(otio.adapters.otio_json.write_to_string(evt))
        self.assertEqual(result["OTIO_SCHEMA"], "EnvVarTestSchema.2")

        # env variable should make a downgrade by default...
        os.environ["OTIO_DEFAULT_TARGET_VERSION_FAMILY_LABEL"] = (
            "TEST_FAMILY_NAME:TEST_LABEL"
        )
        result = json.loads(otio.adapters.otio_json.write_to_string(evt))
        self.assertEqual(result["OTIO_SCHEMA"], "EnvVarTestSchema.1")

        # ...but can still be overridden by passing in an argument
        result = json.loads(otio.adapters.otio_json.write_to_string(evt, {}))
        self.assertEqual(result["OTIO_SCHEMA"], "EnvVarTestSchema.2")

    def test_garbage_env_variables(self):
        cl = otio.schema.Clip()
        invalid_env_error = otio.exceptions.InvalidEnvironmentVariableError

        # missing ":"
        os.environ["OTIO_DEFAULT_TARGET_VERSION_FAMILY_LABEL"] = (
            "invalid_formatting"
        )
        with self.assertRaises(invalid_env_error):
            otio.adapters.otio_json.write_to_string(cl)

        # asking for family/label that doesn't exist in the plugins
        os.environ["OTIO_DEFAULT_TARGET_VERSION_FAMILY_LABEL"] = (
            "nosuch:labelorfamily"
        )
        with self.assertRaises(invalid_env_error):
            otio.adapters.otio_json.write_to_string(cl)

    def test_two_version_manifests(self):
        """test that two manifests layer correctly"""

        fst = otio.plugins.manifest.manifest_from_string(FIRST_MANIFEST)
        snd = otio.plugins.manifest.manifest_from_string(SECOND_MANIFEST)
        fst.extend(snd)

        self.assertIn("UNIQUE_FAMILY", fst.version_manifests)

        lay_fam = fst.version_manifests["LAYERED_FAMILY"]

        self.assertIn("June2022", lay_fam)
        self.assertIn("April2022", lay_fam)
        self.assertEqual(lay_fam["May2022"]["SimpleClass"], 2)


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Test harness for Item."""

import unittest

import opentimelineio as otio
import opentimelineio.test_utils as otio_test_utils

# add Item to the type registry for the purposes of unit testing
# otio.core.register_type(otio.core.Item)


class GapTester(unittest.TestCase, otio_test_utils.OTIOAssertions):

    def test_str_gap(self):
        gp = otio.schema.Gap()
        self.assertMultiLineEqual(
            str(gp),
            "Gap(" +
            str(gp.name) + ", " +
            str(gp.source_range) + ", " +
            str(gp.effects) + ", " +
            str(gp.markers) + ", " +
            str(gp.enabled) + ", " +
            str(gp.metadata) +
            ")"
        )
        self.assertMultiLineEqual(
            repr(gp),
            "otio.schema.Gap("
            "name={}, "
            "source_range={}, "
            "effects={}, "
            "markers={}, "
            "enabled={}, "
            "metadata={}"
            ")".format(
                repr(gp.name),
                repr(gp.source_range),
                repr(gp.effects),
                repr(gp.markers),
                repr(gp.enabled),
                repr(gp.metadata),
            )
        )

        encoded = otio.adapters.otio_json.write_to_string(gp)
        decoded = otio.adapters.otio_json.read_from_string(encoded)
        self.assertJsonEqual(gp, decoded)

    def test_convert_from_filler(self):
        gp = otio.schema.Gap()
        gp._serializable_label = "Filler.1"
        encoded = otio.adapters.otio_json.write_to_string(gp)
        decoded = otio.adapters.otio_json.read_from_string(encoded)
        isinstance(decoded, otio.schema.Gap)

    def test_not_both_source_range_and_duration(self):
        with self.assertRaises(TypeError):
            otio.schema.Gap(
                duration=otio.opentime.RationalTime(10, 24),
                source_range=otio.opentime.TimeRange(
                    otio.opentime.RationalTime(0, 24),
                    otio.opentime.RationalTime(10, 24)
                )
            )

        self.assertJsonEqual(
            otio.schema.Gap(
                duration=otio.opentime.RationalTime(10, 24),
            ),
            otio.schema.Gap(
                source_range=otio.opentime.TimeRange(
                    otio.opentime.RationalTime(0, 24),
                    otio.opentime.RationalTime(10, 24)
                )
            )
        )


class ItemTests(unittest.TestCase, otio_test_utils.OTIOAssertions):

    def test_constructor(self):
        tr = otio.opentime.TimeRange(
            otio.opentime.RationalTime(0, 1),
            otio.opentime.RationalTime(10, 1)
        )
        it = otio.core.Item(name="foo", source_range=tr)
        self.assertEqual(it.source_range, tr)
        self.assertEqual(it.name, "foo")

        encoded = otio.adapters.otio_json.write_to_string(it)
        decoded = otio.adapters.otio_json.read_from_string(encoded)
        self.assertIsOTIOEquivalentTo(it, decoded)

    def test_copy_arguments(self):
        # make sure all the arguments are copied and not referenced
        tr = otio.opentime.TimeRange(
            otio.opentime.RationalTime(0, 24),
            otio.opentime.RationalTime(10, 24),
        )
        name = "foobar"
        effects = []
        markers = []
        metadata = {}
        it = otio.core.Item(
            name=name,
            source_range=tr,
            effects=effects,
            markers=markers,
            metadata=metadata,
        )
        name = 'foobaz'
        self.assertNotEqual(it.name, name)

        tr = otio.opentime.TimeRange(
            otio.opentime.RationalTime(1, tr.start_time.rate),
            duration=tr.duration
        )
        self.assertNotEqual(it.source_range.start_time, tr.start_time)
        markers.append(otio.schema.Marker())
        self.assertNotEqual(it.markers, markers)
        metadata['foo'] = 'bar'
        self.assertNotEqual(it.metadata, metadata)

    def test_duration(self):
        it = otio.core.Item()

        tr = otio.opentime.TimeRange(
            otio.opentime.RationalTime(0, 1),
            otio.opentime.RationalTime(10, 1)
        )
        it = otio.core.Item(source_range=tr)

        self.assertEqual(it.duration(), tr.duration)

    def test_available_range(self):
        it = otio.core.Item()

        with self.assertRaises(NotImplementedError):
            it.available_range()

    def test_duration_and_source_range(self):
        it = otio.core.Item()

        with self.assertRaises(NotImplementedError):
            it.duration()

        self.assertEqual(None, it.source_range)

        tr = otio.opentime.TimeRange(
            otio.opentime.RationalTime(1, 1),
            otio.opentime.RationalTime(10, 1)
        )
        it2 = otio.core.Item(source_range=tr)

        self.assertEqual(tr, it2.source_range)
        self.assertEqual(tr.duration, it2.duration())
        self.assertIsNot(tr.duration, it2.duration())

    def test_trimmed_range(self):
        it = otio.core.Item()
        with self.assertRaises(NotImplementedError):
            it.trimmed_range()
        tr = otio.opentime.TimeRange(
            otio.opentime.RationalTime(1, 1),
            otio.opentime.RationalTime(10, 1)
        )
        it2 = otio.core.Item(source_range=tr)
        self.assertEqual(it2.trimmed_range(), tr)
        self.assertIsNot(it2.trimmed_range(), tr)

    def test_serialize(self):
        tr = otio.opentime.TimeRange(
            otio.opentime.RationalTime(0, 1),
            otio.opentime.RationalTime(10, 1)
        )
        it = otio.core.Item(source_range=tr)
        encoded = otio.adapters.otio_json.write_to_string(it)
        decoded = otio.adapters.otio_json.read_from_string(encoded)
        self.assertIsOTIOEquivalentTo(it, decoded)

    def test_stringify(self):
        tr = otio.opentime.TimeRange(
            duration=otio.opentime.RationalTime(10, 1))
        it = otio.core.Item(source_range=tr)
        self.assertMultiLineEqual(
            str(it),
            "Item("
            "{}, "
            "{}, "
            "{}, "
            "{}, "
            "{}, "
            "{}"
            ")".format(
                str(it.name),
                str(it.source_range),
                str(it.effects),
                str(it.markers),
                str(it.enabled),
                str(it.metadata),
            )
        )

        self.assertMultiLineEqual(
            repr(it),
            "otio.core.Item("
            "name={}, "
            "source_range={}, "
            "effects={}, "
            "markers={}, "
            "enabled={}, "
            "metadata={}"
            ")".format(
                repr(it.name),
                repr(it.source_range),
                repr(it.effects),
                repr(it.markers),
                repr(it.enabled),
                repr(it.metadata),
            )
        )

    def test_metadata(self):
        tr = otio.opentime.TimeRange(
            duration=otio.opentime.RationalTime(10, 1)
        )
        it = otio.core.Item(source_range=tr)
        it.metadata["foo"] = "bar"
        encoded = otio.adapters.otio_json.write_to_string(it)
        decoded = otio.adapters.otio_json.read_from_string(encoded)
        self.assertIsOTIOEquivalentTo(it, decoded)
        self.assertEqual(decoded.metadata["foo"], it.metadata["foo"])

        foo = it.metadata.pop("foo")
        self.assertEqual(foo, "bar")
        foo = it.metadata.pop("foo", "default")
        self.assertEqual(foo, "default")
        with self.assertRaises(KeyError):
            it.metadata.pop("foo")

    def test_add_effect(self):
        tr = otio.opentime.TimeRange(
            duration=otio.opentime.RationalTime(10, 1)
        )
        it = otio.core.Item(source_range=tr)
        it.effects.append(
            otio.schema.Effect(
                effect_name="blur",
                metadata={
                    'amount': '100'
                }
            )
        )
        encoded = otio.adapters.otio_json.write_to_string(it)
        decoded = otio.adapters.otio_json.read_from_string(encoded)
        self.assertIsOTIOEquivalentTo(it, decoded)
        self.assertJsonEqual(it.effects, decoded.effects)

    def test_add_marker(self):
        tr = otio.opentime.TimeRange(
            duration=otio.opentime.RationalTime(10, 1)
        )
        it = otio.core.Item(source_range=tr)
        it.markers.append(
            otio.schema.Marker(
                name="test_marker",
                marked_range=tr,
                metadata={
                    'some stuff to mark': '100'
                }
            )
        )
        encoded = otio.adapters.otio_json.write_to_string(it)
        decoded = otio.adapters.otio_json.read_from_string(encoded)
        self.assertIsOTIOEquivalentTo(it, decoded)
        self.assertJsonEqual(it.markers, decoded.markers)

    def test_enabled(self):
        tr = otio.opentime.TimeRange(
            duration=otio.opentime.RationalTime(10, 1)
        )
        it = otio.core.Item(source_range=tr)
        self.assertEqual(it.enabled, True)
        it.enabled = False
        self.assertEqual(it.enabled, False)
        encoded = otio.adapters.otio_json.write_to_string(it)
        decoded = otio.adapters.otio_json.read_from_string(encoded)
        self.assertIsOTIOEquivalentTo(it, decoded)
        self.assertJsonEqual(it.enabled, decoded.enabled)

    def test_copy(self):
        tr = otio.opentime.TimeRange(
            duration=otio.opentime.RationalTime(10, 1)
        )
        it = otio.core.Item(source_range=tr, metadata={"foo": "bar"})
        it.markers.append(
            otio.schema.Marker(
                name="test_marker",
                marked_range=tr,
                metadata={
                    'some stuff to mark': '100'
                }
            )
        )
        it.effects.append(
            otio.schema.Effect(
                effect_name="blur",
                metadata={
                    'amount': '100'
                }
            )
        )

        it.metadata["foo"] = "bar2"

        # deep copy should have different dictionaries
        it_dcopy = it.deepcopy()
        it_dcopy.metadata["foo"] = "not bar"
        self.assertNotEqual(it.metadata, it_dcopy.metadata)

    def test_copy_library(self):
        tr = otio.opentime.TimeRange(
            duration=otio.opentime.RationalTime(10, 1)
        )
        it = otio.core.Item(source_range=tr, metadata={"foo": "bar"})
        it.markers.append(
            otio.schema.Marker(
                name="test_marker",
                marked_range=tr,
                metadata={
                    'some stuff to mark': '100'
                }
            )
        )
        it.effects.append(
            otio.schema.Effect(
                effect_name="blur",
                metadata={
                    'amount': '100'
                }
            )
        )

        import copy

        # deep copy should have different dictionaries
        it_dcopy = copy.deepcopy(it)
        it_dcopy.metadata["foo"] = "not bar"
        self.assertNotEqual(it.metadata, it_dcopy.metadata)

    def test_visible_range(self):
        timeline = otio.schema.Timeline(
            tracks=[
                otio.schema.Track(
                    name="V1",
                    children=[
                        otio.schema.Clip(
                            name="A",
                            source_range=otio.opentime.TimeRange(
                                start_time=otio.opentime.RationalTime(
                                    value=1,
                                    rate=30
                                ),
                                duration=otio.opentime.RationalTime(
                                    value=50,
                                    rate=30
                                )
                            )
                        ),
                        otio.schema.Transition(
                            in_offset=otio.opentime.RationalTime(
                                value=7,
                                rate=30
                            ),
                            out_offset=otio.opentime.RationalTime(
                                value=10,
                                rate=30
                            ),
                        ),
                        otio.schema.Clip(
                            name="B",
                            source_range=otio.opentime.TimeRange(
                                start_time=otio.opentime.RationalTime(
                                    value=100,
                                    rate=30
                                ),
                                duration=otio.opentime.RationalTime(
                                    value=50,
                                    rate=30
                                )
                            )
                        ),
                        otio.schema.Transition(
                            in_offset=otio.opentime.RationalTime(
                                value=17,
                                rate=30
                            ),
                            out_offset=otio.opentime.RationalTime(
                                value=15,
                                rate=30
                            ),
                        ),
                        otio.schema.Clip(
                            name="C",
                            source_range=otio.opentime.TimeRange(
                                start_time=otio.opentime.RationalTime(
                                    value=50,
                                    rate=30
                                ),
                                duration=otio.opentime.RationalTime(
                                    value=50,
                                    rate=30
                                )
                            )
                        ),
                        otio.schema.Clip(
                            name="D",
                            source_range=otio.opentime.TimeRange(
                                start_time=otio.opentime.RationalTime(
                                    value=1,
                                    rate=30
                                ),
                                duration=otio.opentime.RationalTime(
                                    value=50,
                                    rate=30
                                )
                            )
                        )
                    ]
                )
            ]
        )
        self.maxDiff = None
        self.assertListEqual(
            ["A", "B", "C", "D"],
            [item.name for item in timeline.each_clip()]
        )
        self.assertListEqual(
            [
                otio.opentime.TimeRange(
                    start_time=otio.opentime.RationalTime(
                        value=1,
                        rate=30
                    ),
                    duration=otio.opentime.RationalTime(
                        value=50,
                        rate=30
                    )
                ),
                otio.opentime.TimeRange(
                    start_time=otio.opentime.RationalTime(
                        value=100,
                        rate=30
                    ),
                    duration=otio.opentime.RationalTime(
                        value=50,
                        rate=30
                    )
                ),
                otio.opentime.TimeRange(
                    start_time=otio.opentime.RationalTime(
                        value=50,
                        rate=30
                    ),
                    duration=otio.opentime.RationalTime(
                        value=50,
                        rate=30
                    )
                ),
                otio.opentime.TimeRange(
                    start_time=otio.opentime.RationalTime(
                        value=1,
                        rate=30
                    ),
                    duration=otio.opentime.RationalTime(
                        value=50,
                        rate=30
                    )
                ),
            ],
            [item.trimmed_range() for item in timeline.each_clip()]
        )

        self.assertListEqual(
            [
                otio.opentime.TimeRange(
                    start_time=otio.opentime.RationalTime(
                        value=1,
                        rate=30
                    ),
                    duration=otio.opentime.RationalTime(
                        value=50 + 10,
                        rate=30
                    )
                ),
                otio.opentime.TimeRange(
                    start_time=otio.opentime.RationalTime(
                        value=100 - 7,
                        rate=30
                    ),
                    duration=otio.opentime.RationalTime(
                        value=50 + 15 + 7,
                        rate=30
                    )
                ),
                otio.opentime.TimeRange(
                    start_time=otio.opentime.RationalTime(
                        value=33,
                        rate=30
                    ),
                    duration=otio.opentime.RationalTime(
                        value=50 + 17,
                        rate=30
                    )
                ),
                otio.opentime.TimeRange(
                    start_time=otio.opentime.RationalTime(
                        value=1,
                        rate=30
                    ),
                    duration=otio.opentime.RationalTime(
                        value=50,
                        rate=30
                    )
                ),
            ],
            [item.visible_range() for item in timeline.each_clip()]
        )


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Unit tests for the 'examples'"""
import unittest
import sys
import os
import subprocess


# handle python2 vs python3 difference
try:
    from tempfile import TemporaryDirectory  # noqa: F401
    import tempfile
except ImportError:
    # XXX: python2.7 only
    from backports import tempfile

import opentimelineio as otio


class BuildSimpleTimelineExampleTest(unittest.TestCase):
    """use the build_simple_timeline.py example to generate timelines"""

    def test_duration(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_file = os.path.join(temp_dir, "test_basic.otio")

            examples_path = os.path.join(
                os.path.dirname(os.path.dirname(__file__)),
                "examples",
                "build_simple_timeline.py",
            )

            subprocess.check_call(
                [sys.executable, examples_path, temp_file],
                stdout=subprocess.PIPE
            )
            known = otio.adapters.read_from_file(temp_file)

            # checks against a couple of the adapters
            for suffix in [".xml", ".edl", ".otio"]:
                this_test_file = temp_file.replace(".otio", suffix)
                subprocess.check_call(
                    [sys.executable, examples_path, this_test_file],
                    stdout=subprocess.PIPE
                )
                test_result = otio.adapters.read_from_file(this_test_file)
                self.assertEqual(known.duration(), test_result.duration())


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

import unittest
import os

import opentimelineio as otio
from tests import baseline_reader, utils


# handle python2 vs python3 difference
try:
    from tempfile import TemporaryDirectory  # noqa: F401
    import tempfile
except ImportError:
    # XXX: python2.7 only
    from backports import tempfile


"""Unit tests for the adapter plugin system."""


MANIFEST_PATH = "adapter_plugin_manifest.plugin_manifest"
ADAPTER_PATH = "adapter_example"
MEDIA_LINKER_EXAMPLE = "media_linker_example"


class TestAdapterSuffixes(unittest.TestCase):
    def test_supported_suffixes_is_not_none(self):
        result = otio.adapters.suffixes_with_defined_adapters()
        self.assertIsNotNone(result)
        self.assertNotEqual(result, [])
        self.assertNotEqual(result, set([]))


class TestPluginAdapters(unittest.TestCase):
    def setUp(self):
        self.jsn = baseline_reader.json_baseline_as_string(ADAPTER_PATH)
        self.adp = otio.adapters.read_from_string(self.jsn, 'otio_json')
        self.adp._json_path = os.path.join(
            baseline_reader.MODPATH,
            "baselines",
            ADAPTER_PATH
        )

    def test_plugin_adapter(self):
        self.assertEqual(self.adp.name, "example")
        self.assertEqual(self.adp.execution_scope, "in process")
        self.assertEqual(self.adp.filepath, "example.py")
        self.assertEqual(self.adp.suffixes[0], u"example")
        self.assertEqual(list(self.adp.suffixes), [u'example'])

        self.assertMultiLineEqual(
            str(self.adp),
            "Adapter("
            "{}, "
            "{}, "
            "{}, "
            "{}"
            ")".format(
                repr(self.adp.name),
                repr(self.adp.execution_scope),
                repr(self.adp.filepath),
                repr(self.adp.suffixes),
            )
        )
        self.assertMultiLineEqual(
            repr(self.adp),
            "otio.adapter.Adapter("
            "name={}, "
            "execution_scope={}, "
            "filepath={}, "
            "suffixes={}"
            ")".format(
                repr(self.adp.name),
                repr(self.adp.execution_scope),
                repr(self.adp.filepath),
                repr(self.adp.suffixes),
            )
        )

        self.assertNotEqual(self.adp._json_path, None)

    def test_load_adapter_module(self):
        target = os.path.join(
            baseline_reader.MODPATH,
            "baselines",
            "example.py"
        )

        self.assertEqual(self.adp.module_abs_path(), target)
        self.assertTrue(hasattr(self.adp.module(), "read_from_file"))

        # call through the module accessor
        self.assertEqual(self.adp.module().read_from_file("foo").name, "foo")

        # call through the convienence wrapper
        self.assertEqual(self.adp.read_from_file("foo").name, "foo")

    def test_has_feature(self):
        self.assertTrue(self.adp.has_feature("read"))
        self.assertTrue(self.adp.has_feature("read_from_file"))
        self.assertFalse(self.adp.has_feature("write"))

    def test_pass_arguments_to_adapter(self):
        self.assertEqual(self.adp.read_from_file("foo", suffix=3).name, "foo3")

    def test_run_media_linker_during_adapter(self):
        mfest = otio.plugins.ActiveManifest()

        manifest = utils.create_manifest()
        # this wires up the media linkers into the active manifest
        mfest.media_linkers.extend(manifest.media_linkers)
        fake_tl = self.adp.read_from_file("foo", media_linker_name="example")

        self.assertTrue(
            fake_tl.tracks[0][0].media_reference.metadata.get(
                'from_test_linker'
            )
        )

        fake_tl = self.adp.read_from_string(
            "foo",
            media_linker_name="example"
        )

        self.assertTrue(
            fake_tl.tracks[0][0].media_reference.metadata.get(
                'from_test_linker'
            )
        )

        # explicitly turn the media_linker off
        fake_tl = self.adp.read_from_file("foo", media_linker_name=None)
        self.assertIsNone(
            fake_tl.tracks[0][0].media_reference.metadata.get(
                'from_test_linker'
            )
        )

        # Delete the temporary manifest
        utils.remove_manifest(manifest)


class TestPluginManifest(unittest.TestCase):

    def setUp(self):
        self.man = utils.create_manifest()

    def tearDown(self):
        utils.remove_manifest(self.man)

    def test_plugin_manifest(self):
        self.assertNotEqual(self.man.adapters, [])

    def test_find_adapter_by_suffix(self):
        self.assertEqual(self.man.from_filepath("example").name, "example")
        with self.assertRaises(Exception):
            self.man.from_filepath("BLARG")
        adp = self.man.from_filepath("example")
        self.assertEqual(adp.module().read_from_file("path").name, "path")
        self.assertEqual(
            self.man.adapter_module_from_suffix(
                "example"
            ).read_from_file("path").name,
            "path"
        )

    def test_find_adapter_by_name(self):
        self.assertEqual(self.man.from_name("example").name, "example")
        with self.assertRaises(Exception):
            self.man.from_name("BLARG")
        adp = self.man.from_name("example")
        self.assertEqual(adp.module().read_from_file("path").name, "path")
        self.assertEqual(
            self.man.adapter_module_from_name("example").read_from_file(
                "path"
            ).name,
            "path"
        )

    def test_deduplicate_env_variable_paths(self):
        "Ensure that duplicate entries in the environment variable are ignored"

        basename = "unittest.plugin_manifest.json"

        # back up existing manifest
        bak = otio.plugins.manifest._MANIFEST
        bak_env = os.environ.get('OTIO_PLUGIN_MANIFEST_PATH')

        try:
            # Generate a fake manifest in a temp file, and point at it with
            # the environment variable
            with tempfile.TemporaryDirectory(
                prefix='test_find_manifest_by_environment_variable'
            ) as temp_dir:
                temp_file = os.path.join(temp_dir, basename)
                otio.adapters.write_to_file(self.man, temp_file, 'otio_json')

                # clear out existing manifest
                otio.plugins.manifest._MANIFEST = None

                # set where to find the new manifest
                os.environ['OTIO_PLUGIN_MANIFEST_PATH'] = (
                    temp_file
                    # add it twice
                    + os.pathsep + temp_file
                )
                result = otio.plugins.manifest.load_manifest()

                # ... should only appear once in the result
                self.assertEqual(result.source_files.count(temp_file), 1)

                # Rather than try and remove any other setuptools based plugins
                # that might be installed, this check is made more permissive to
                # see if the known unit test linker is being loaded by the manifest
                self.assertTrue(len(result.media_linkers) > 0)
                self.assertIn("example", (ml.name for ml in result.media_linkers))

        finally:
            otio.plugins.manifest._MANIFEST = bak
            if bak_env is not None:
                os.environ['OTIO_PLUGIN_MANIFEST_PATH'] = bak_env
            else:
                if "OTIO_PLUGIN_MANIFEST_PATH" in os.environ:
                    del os.environ['OTIO_PLUGIN_MANIFEST_PATH']

    def test_find_manifest_by_environment_variable(self):
        basename = "unittest.plugin_manifest.json"

        # back up existing manifest
        bak = otio.plugins.manifest._MANIFEST
        bak_env = os.environ.get('OTIO_PLUGIN_MANIFEST_PATH')

        try:
            # Generate a fake manifest in a temp file, and point at it with
            # the environment variable
            with tempfile.TemporaryDirectory(
                prefix='test_find_manifest_by_environment_variable'
            ) as temp_dir:
                temp_file = os.path.join(temp_dir, basename)
                otio.adapters.write_to_file(self.man, temp_file, 'otio_json')

                # clear out existing manifest
                otio.plugins.manifest._MANIFEST = None

                # set where to find the new manifest
                os.environ['OTIO_PLUGIN_MANIFEST_PATH'] = (
                    temp_file + os.pathsep + 'foo'
                )
                result = otio.plugins.manifest.load_manifest()

                # Rather than try and remove any other setuptools based plugins
                # that might be installed, this check is made more permissive to
                # see if the known unit test linker is being loaded by the manifest
                self.assertTrue(len(result.media_linkers) > 0)
                self.assertIn("example", (ml.name for ml in result.media_linkers))
        finally:
            otio.plugins.manifest._MANIFEST = bak
            if bak_env is not None:
                os.environ['OTIO_PLUGIN_MANIFEST_PATH'] = bak_env
            else:
                del os.environ['OTIO_PLUGIN_MANIFEST_PATH']

    def test_load_manifest_empty_environment_variable(self):
        # back up existing manifest
        bak = otio.plugins.manifest._MANIFEST
        bak_env = os.environ.get('OTIO_PLUGIN_MANIFEST_PATH')

        try:
            # clear out existing manifest
            otio.plugins.manifest._MANIFEST = None

            # set manifest to an empty path
            os.environ['OTIO_PLUGIN_MANIFEST_PATH'] = ''
            result = otio.plugins.manifest.load_manifest()

            # Make sure adapters and linkers landed in the proper place
            for adapter in result.adapters:
                self.assertIsInstance(adapter, otio.adapters.Adapter)

            for linker in result.media_linkers:
                self.assertIsInstance(linker, otio.media_linker.MediaLinker)
        finally:
            otio.plugins.manifest._MANIFEST = bak
            if bak_env is not None:
                os.environ['OTIO_PLUGIN_MANIFEST_PATH'] = bak_env
            else:
                del os.environ['OTIO_PLUGIN_MANIFEST_PATH']

    def test_plugin_manifest_order(self):
        basename = "test.plugin_manifest.json"

        # back up existing manifest
        bak = otio.plugins.manifest._MANIFEST
        bak_env = os.environ.get('OTIO_PLUGIN_MANIFEST_PATH')
        try:
            local_manifest = {
                "OTIO_SCHEMA": "PluginManifest.1",
                "adapters": [
                    {
                        "OTIO_SCHEMA": "Adapter.1",
                        "name": "local_json",
                        "execution_scope": "in process",
                        "filepath": "example.py",
                        "suffixes": ["example"]
                    }
                ],
            }

            with tempfile.TemporaryDirectory() as temp_dir:
                filename = os.path.join(temp_dir, basename)

                otio.adapters.write_to_file(local_manifest, filename, 'otio_json')

                result = otio.plugins.manifest.load_manifest()
                self.assertTrue(len(result.adapters) > 0)
                self.assertIn("otio_json", (ml.name for ml in result.adapters))
                self.assertNotIn("local_otio", (ml.name for ml in result.adapters))

                # set where to find the new manifest
                os.environ['OTIO_PLUGIN_MANIFEST_PATH'] = filename
                result = otio.plugins.manifest.load_manifest()

                # Rather than try and remove any other setuptools based plugins
                # that might be installed, this check is made more permissive to
                # see if the known unit test linker is being loaded by the manifest
                self.assertTrue(len(result.adapters) > 0)
                self.assertIn("otio_json", (ml.name for ml in result.adapters))
                self.assertIn("local_json", (ml.name for ml in result.adapters))
                self.assertLess(
                    [ml.name for ml in result.adapters].index("local_json"),
                    [ml.name for ml in result.adapters].index("otio_json")
                )
        finally:
            otio.plugins.manifest._MANIFEST = bak
            if bak_env is not None:
                os.environ['OTIO_PLUGIN_MANIFEST_PATH'] = bak_env
            else:
                del os.environ['OTIO_PLUGIN_MANIFEST_PATH']


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

import pkg_resources

from opentimelineio.plugins import manifest

"""An example plugin package that generates its package manifest on demand.

If you create a plugin that doesn't have a plugin_manifest.json, OTIO will
attempt to call the plugin_manifest() function from the __init__.py directory.

This would allow you to programmatically generate a manifest rather than have
it be static on disk, allowing you to switch features on or off or do some
template substition or any other kind of procedural processing.

This unit test uses a very simple example that just reads the manifest from
a non-standard json file path.
"""


def plugin_manifest():
    # XXX: in python3.5 resource_string returns a 'bytes' object, but
    #      json.loads requires a string, not bytes (only in 3.5 -- 2.7 and 3.6
    #      seem to both be working).  Luckily .decode() seems to work in both
    #      python3 and python2, so this *should* work for both versions.
    return manifest.manifest_from_string(
        pkg_resources.resource_string(
            __name__,
            'unusually_named_plugin_manifest.json'
        ).decode('utf-8')
    )
